#ifndef NESTARRAY_CPP
#define NESTARRAY_CPP
#ifndef NESTARRAY_H
  #include "nestarray.h"
#endif

/****************************************************************************/
// PURPOSE:
//   Linear position to matrix coordinates conversion function using
//   calulated linear offsets vector to determine results.
//
// PRE:
//   int Pos_      : Linear position value to be converted into data matrix
//                   coordinates.
//   int* Coords_  : int vector to be filled with the data matrix coordinates.
//   int CoordLen_ : Length of the data matrix coordinates vector excluding
//                   zero index element.
//   int* OffSets_ : Progressive linear position offsets to be used in
//                   continuous divisions of the linear position value to
//                   give the equivalent data matrix coordinates.
//
//   int OffsetsLen_ : Length of the linear offsets vector.
//
// POST:
//   Returns the int vector that is filled with the data matrix coordinates
//   that is equivalent to the specified linear position value.
//
int* LinearToCoords(int Pos_, int* Coords_, int CoordsLen_,
                    int* OffSets_, int OffSetsLen_)
{
  int y;
  int x = 1;

  if (Pos_ / OffSets_[x])
    return NULL;

  x += 2;
  for (y = 1; y < CoordsLen_ && x < OffSetsLen_; y++)
  {
    Coords_[y] = Pos_ / OffSets_[x];
    Pos_ %= OffSets_[x];
    x += 2;
  }
  
  if (y == CoordsLen_)
  {  
    Coords_[0] = CoordsLen_;
    Coords_[CoordsLen_] = Pos_;
    return Coords_;
  }

  return NULL;
}

/****************************************************************************/
// PURPOSE:
//   Matrix coordinates to linear position conversion function using
//   calculated linear vector offsets to determine results.
//
// PRE:
//   int* Coords_    : Data matrix coordinates.
//   int CoordsLen_  : Length of the data matrix coordinates vector excluding
//                     zero index element.
//   int* Offsets_   : Progressive linear position offsets to be multiplied
//                     with the data matrix coordinates to give the
//                     equivalent linear position value.
//   int OffSetsLen_ : Length of the linear position offsets vector.
//
// POST;
//   Returns the equivalent linear position value for the given data matrix
//   coordinates.
//
int CoordsToLinear(int* Coords_, int CoordsLen_,
                   int* OffSets_, int OffSetsLen_)
{
  int y;
  int x = 2;  
  int pos = 0;

  for (y = 1; y < CoordsLen_ && x < OffSetsLen_ &&
       InRangeX(-1, Coords_[y], OffSets_[x++]); y++)
    pos += (Coords_[y] * OffSets_[x++]);
  pos += Coords_[CoordsLen_];
  
  return ((y == CoordsLen_) ? pos:-1);
}

/****************************************************************************/
// PURPOSE:
//   Method to calculate and return an allocated int array of linear
//   position offsets given a int vector filled with the array lengths of
//   the constituent arrays within the data matrix.
//
// Offset calculation scheme example:
//
//   [asz1][cnt1]
//   [asz2][cnt2==cnt1*asz1]
//   [asz3][cnt3==cnt2*asz2]
//   [asz4][cnt4==cnt3*asz3]
//   [asz5][cnt5==cnt4*asz4]
//   [asz6][cnt6==cnt5*asz5]
//   [ 1  ][cnt7==cnt6*asz6]
//
// Created offset array in reverse order
// [asz1][cnt1][asz2][cnt2][asz3][cnt3][asz4][cnt4][asz5][cnt5][asz6][cnt6]
// [ 1  ][cnt7]
//
// Resorted to go from largest base matrix vector to smallest element vector
// [ 1  ][cnt7][asz6][cnt6][asz5][cnt5][asz4][cnt4][asz3][cnt3][asz2][cnt2]
// [asz1][cnt1]
//
// Matrix position coordinates scheme example:
//
//   PosVect[0] = (Not Used or Set With Position Vector Length)
//   PosVect[1] = Holds index position into 2nd level array
//   PosVect[2] = Holds index position into 3rd level array
//   ...
//   PosVect[Dim] = Holds index into basic array element data in 1d vector
//
// PRE:
//   int* ArraySizes_ : The array lengths of the constituent arrays within
//                      the data matrix.
//   int ArrayLen_    : The length of the int vector containing the array
//                      lengths.
//
// POST:
//   Returns an int array filled with the linear position offsets of the
//   constituent arrays within the data matrix.
//
int* CalculateOffsets(int* ArraySizes_, int ArrayLen_, int& OffSetsLen_)
{
  if (!ArraySizes_)
    return NULL;

  int x;
  int i = 1;
  int cnt = 1;
  int* OffSets_ = (int*)RawAllocateWith(MEMMATRIX, sizeof(int) * ArrayLen_ * 2);

  // Working backwards from array length of deepest 1d element to top level
  // container matrix length.
  for (x = ArrayLen_ - 1; x >= 0; x--)
  {
    cnt *= ArraySizes_[x];
    OffSets_[i] = cnt;
    OffSets_[i-1] = x ? ArraySizes_[x-1]:1;
    i += 2;
  }

  OffSetsLen_ = i - 1;
  i -= 3;
  x = 0;

  // Resorting offest array to go from largest base matrix vector to
  // smallest element vector.
  while (x < i)
  {
    cnt = OffSets_[x];
    OffSets_[x++] = OffSets_[i];
    OffSets_[i++] = cnt;

    cnt = OffSets_[x];
    OffSets_[x++] = OffSets_[i];
    OffSets_[i++] = cnt;
    i -= 4;
  }

  return OffSets_;
}

/****************************************************************************/
// PURPOSE:
//   Method to initialize individual vectors withing the data matrix.
//
// PRE:
//   void* Array_  : Indexed constituent array within the data matrix.
//   int Dim_      : The dimensions of the matrix, value range = [3...8]
//   int Off_      : The offset into the data matrix, value range = [0...5]
//   int Index_    : Index accessing constituent array of the data matrix,
//                   value range = [1...7]
//   int* LenVect_ : Length vector containing array lengths of constituent
//                   vectors from the general data matrix.
//
// POST:
//   Returns the indexed array after it has been initiallzed.
//
void* InitMatrix(void* Array_, int Dim_, int Off_, int Index_, int* LenVect_)
{
  int x;
  int sz = SizeOfVector(Dim_-Off_, Index_-Off_, LenVect_);
  
  int******** Vect8_ = (Index_ == 0) ? (int********)Array_:NULL;
  int******* Vect7_ = (Index_ == 1) ? (int*******)Array_:NULL;
  int****** Vect6_ = (Index_ == 2) ? (int******)Array_:NULL;
  int***** Vect5_ = (Index_ == 3) ? (int*****)Array_:NULL;
  int**** Vect4_ = (Index_ == 4) ? (int****)Array_:NULL;
  int*** Vect3_ = (Index_ == 5) ? (int***)Array_:NULL;
  int** Vect2_ = (Index_ == 6) ? (int**)Array_:NULL;
  int* Vect1_ = (Index_ == 7) ? (int*)Array_:NULL;
  
  for (x = 1; x < sz; x++)
  {
    switch (Index_)
    {
      case 0:
        Vect8_[x] = NULL;
        break;
        
      case 1:
        Vect7_[x] = NULL;
        break;
    
      case 2:
        Vect6_[x] = NULL;
        break;
    
      case 3:
        Vect5_[x] = NULL;
        break;
        
      case 4:
        Vect4_[x] = NULL;
        break;

      case 5:
        Vect3_[x] = NULL;
        break;

      case 6:
        Vect2_[x] = NULL;
        break;

      case 7:
        Vect1_[x] = 0;
    }    
  }

  return Array_;
}

/****************************************************************************/
// PURPOSE:
//   Method to set the parent references vector from the general data matrix
//   with values from a position vector indexed with a parent index value
//
// array[1...n]
//      [i=2...n][0][0][0][0] = 3rdArr array size=x index=i=2 of
//                              2nd...nth 2ndArr element. 0 = deleted.
//
// array[1...n]
//      [i=2...n][0][0][0][(i+1)=3...x+i+1] = Deallocated array flag = 0/1.
//                                          = Shallow array copy flag = 0/2.
//
// array[1...n]
//      [i=2...n][1][0][0][(i+1)=3...x+i+1] = 3rdArr to TopPar pointer
//      reference number, referencing TopPar pointer in 1st...nth TopArr
//      element for 3rdArr arrays from 3...x.
//
// array[1...n]
//      [i=2...n][i=2][0][0][(i+1)=3...x+i+1] = 3rdArr to 2ndArr pointer
//      reference number, referencing 2ndArr pointer in 2nd...nth 2ndArr
//      element for 3rdArr arrays from 3...x.
//
// array[1...n]
//      [i=2...n][(i+1)=3...x+i+1]* = 1st starting 3rdArr value element to
//                                    ending 3rdArr value element x.
//
// PRE:
//   int* Vector_     : Parent reference vector retrieved from the data matrix
//   int Dim_         : The dimensions of the matrix, value range = [3...8]
//   int ParentIndex_ : Index into the position vector containing parent
//                      reference values, denoted by variable i = 1...Index_.
//   int Index_       : Index level into position vector, denoted by constant
//                      i == Index_, max = Index_ + 1
//   int* LenVect_    : Length vector containing array lengths of constituent
//                      vectors from the general data matrix.
//   int* PosVect_    : Position vector containing array element positions of
//                      constituent vectros from the general data matrix.
//   bool& Error_     : Error in method if set TRUE, otherwise set to FALSE.
//
int* SetParentReference(int* Vector_, int Dim_,
                        int ParentIndex_, int Index_,
                        int* LenVect_, int* PosVect_, bool& Error_)
{
  int pos;
  int x = Index_;        // x = max - 1

  if (InRange(x+1, PosVect_[x+1], SizeOfVector(Dim_, x, LenVect_) - 1))
    pos = PosVect_[x+1]; // = [(i+1)=3...x+i+1]
  else
    pos = 0;

  Error_ = !pos;
  
  if (pos)               // array[n]...[(i+1)=3...x+i+1] = posvect[1...i]
    Vector_[pos] = PosVect_[ParentIndex_];

  return Vector_;
}

/****************************************************************************/
// PURPOSE:
//   Method to get the parent references vector from the index into the data
//   matrix.
//
// array[1...n]
//      [i=2...n][0][0][0][0] = 3rdArr array size=x index=i=2 of
//                              2nd...nth 2ndArr element. 0 = deleted.
//
// array[1...n]
//      [i=2...n][0][0][0][(i+1)=3...x+i+1] = Deallocated array flag = 0/1.
//                                          = Shallow array copy flag = 0/2.
//
// array[1...n]
//      [i=2...n][1][0][0][(i+1)=3...x+i+1] = 3rdArr to TopPar pointer
//      reference number, referencing TopPar pointer in 1st...nth TopArr
//      element for 3rdArr arrays from 3...x.
//
// array[1...n]
//      [i=2...n][i=2][0][0][(i+1)=3...x+i+1] = 3rdArr to 2ndArr pointer
//      reference number, referencing 2ndArr pointer in 2nd...nth 2ndArr
//      element for 3rdArr arrays from 3...x.
//
// array[1...n]
//      [i=2...n][(i+1)=3...x+i+1]* = 1st starting 3rdArr value element to
//                                    ending 3rdArr value element x.
//
// PRE:
//   int* ParentRefs_ : Parent references vector of specified constituent
//                      vector within general data matrix accessed by
//                      MatrixIndex_ argument
//   int Dim_       : The dimensions of the matrix, value range = [3...8]
//   int Off_       : The offset into the data matrix, value range = [0...5]
//   int MatrixIndex_ : Index accessing constituent array of the data matrix,
//                      value range = [1...7]
//   int* LenVect_  : Length vector containing length of constituent arrays
//   int* PosVect_  : Position vector containing coordinates into matrix
//   void* Array_     : vector retrieved from the data matrix containing
//                      parent array references to be retrieved.
//   bool& Error_   : Error in array indexing (out of range)
//
// POST:
//   Returns parent references vector.
//   Parent reference vector format:
//       ParentRefs[0] == Size of array starting for index element 1
//       ParentRefs[1] == referencing TopPar pointer in 1st...nth TopArr
//       ParentRefs[2] == referencing 2ndArr pointer in 2nd...nth 2ndArr
//       ...
//       ParentRefs[3] == referencing Dim-1 Arr pointer in Dim-1...nth Array
//
int* GetParentReference(int* ParentRefs_, int Dim_, int Off_, int MatrixIndex_,
                        int* LenVect_, int* PosVect_, void* Array_, bool& Error_)
{
  int pos;
  int y;
  int x = MatrixIndex_;
  int xOff_ = x - Off_;
  int xDim_ = Dim_ - Off_;
  int* Vector_;

  int******* Vect7_ = (x == 1) ? (int*******)Array_:NULL;
  int****** Vect6_ = (x == 2) ? (int******)Array_:NULL;
  int***** Vect5_ = (x == 3) ? (int*****)Array_:NULL;
  int**** Vect4_ = (x == 4) ? (int****)Array_:NULL;
  int*** Vect3_ = (x == 5) ? (int***)Array_:NULL;
  int** Vect2_ = (x == 6) ? (int**)Array_:NULL;
  int* Vect1_ = (x == 7) ? (int*)Array_:NULL;

  if (InRange(xOff_+1, PosVect_[xOff_+1], SizeOfVector(xDim_, xOff_, LenVect_) - 1))
    pos = PosVect_[xOff_+1];   // = [(i+1)=3...x+i+1]
  else
    pos = 0;

  Error_ = !pos;  
    
  if (pos)
  {    
    // 0th index of parent reference array is the size of the array
    ParentRefs_[0] = xOff_;
    
    for (y = 1; y <= xOff_; y++)
    {
      // ParentRefs[1...n] = array[1...n]...[(i+1)=3...x+i+1]
      Vector_ =
      (x == 1) ? Vect7_[y][0][0][0][0][0]:
      (x == 2) ? Vect6_[y][0][0][0][0]:      
      (x == 3) ? Vect5_[y][0][0][0]:
      (x == 4) ? Vect4_[y][0][0]:
      (x == 5) ? Vect3_[y][0]:
      (x == 6) ? Vect2_[y]:
      (x == 7) ? Vect1_:NULL;

      if (Vector_)
        ParentRefs_[y] = Vector_[pos];
    }
  }

  return ParentRefs_;
}

/****************************************************************************/
// PURPOSE:
//   Method to set the reference array for each constituent array
//
// array[1...n]
//      [i=2...n][0][0][0][0] = 3rdArr array size=x index=i=2 of
//                              2nd...nth 2ndArr element. 0 = deleted.
//
// array[1...n]
//      [i=2...n][0][0][0][(i+1)=3...x+i+1] = Deallocated array flag = 0/1.
//                                          = Shallow array copy flag = 0/2.
//
// array[1...n]
//      [i=2...n][1][0][0][(i+1)=3...x+i+1] = 3rdArr to TopPar pointer
//      reference number, referencing TopPar pointer in 1st...nth TopArr
//      element for 3rdArr arrays from 3...x.
//
// array[1...n]
//      [i=2...n][i=2][0][0][(i+1)=3...x+i+1] = 3rdArr to 2ndArr pointer
//      reference number, referencing 2ndArr pointer in 2nd...nth 2ndArr
//      element for 3rdArr arrays from 3...x.
//
// array[1...n]
//      [i=2...n][(i+1)=3...x+i+1]* = 1st starting 3rdArr value element to
//                                    ending 3rdArr value element x.
//
// PRE:
//   void* Array   : Data matrix
//   int Dim_      : The dimensions of the matrix, value range = [3...8]
//   int Off_      : The offset into the data matrix, value range = [0...5]
//   int Index_    : Index accessing constituent array of the data matrix,
//                   value range = [1...6]
//   int* LenVect_ : Length vector containing length of constituent arrays
//   int* PosVect_ : Position vector containing coordinates into matrix
//   bool& Error_  : Error in array indexing (out of range)
//
// POST:
//   Returns initialized data matrix matrix of given dimensions of length
//
void* SetRefArrays(void* Array_, int Dim_, int Off_, int Index_,
                   int* LenVect_, int* PosVect_, bool& Error_)
{
  int x, y;
  int sz = SizeOfVector(Dim_-Off_, Index_-Off_, LenVect_);

  int******* Vect7_ = (Index_ == 1) ? (int*******)Array_:NULL;
  int****** Vect6_ = (Index_ == 2) ? (int******)Array_:NULL;
  int***** Vect5_ = (Index_ == 3) ? (int*****)Array_:NULL;
  int**** Vect4_ = (Index_ == 4) ? (int****)Array_:NULL;
  int*** Vect3_ = (Index_ == 5) ? (int***)Array_:NULL;
  int** Vect2_ = (Index_ == 6) ? (int**)Array_:NULL;
  int* Vect1_ = (Index_ == 7) ? (int*)Array_:NULL;

  for (x = Off_+1; x <= Index_; x++)
  {
    // array[1]...array[n]
    switch (Index_)
    {
      case 1:
        Vect6_ = Vect7_[x-Off_] =
        (int******)RawAllocateWith(MEMMATRIX, sizeof(int*****));
        break;
        
      case 2:
        Vect5_ = Vect6_[x-Off_] =
        (int*****)RawAllocateWith(MEMMATRIX, sizeof(int****));
        break;
    
      case 3:
        Vect4_ = Vect5_[x-Off_] =
        (int****)RawAllocateWith(MEMMATRIX, sizeof(int***));
        break;
        
      case 4:
        Vect3_ = Vect4_[x-Off_] =
        (int***)RawAllocateWith(MEMMATRIX, sizeof(int**));
        break;

      case 5:
        Vect2_ = Vect3_[x-Off_] =
        (int**)RawAllocateWith(MEMMATRIX, sizeof(int*));
        break;

      case 6:
        Vect1_ = Vect2_[x-Off_] =
        (int*)RawAllocateWith(MEMMATRIX, sizeof(int) * sz);
    }
  
    for (y = Index_ + 2; y <= Dim_; y++)
    {
      // array[n]...[0]...[0]
      switch (y)
      {
        case 3:
          Vect5_ = Vect6_[0] =
          (int*****)RawAllocateWith(MEMMATRIX, sizeof(int****));
          break;
          
        case 4:
          Vect4_ = Vect5_[0] =
          (int****)RawAllocateWith(MEMMATRIX, sizeof(int***));
          break;
      
        case 5:
          Vect3_ = Vect4_[0] =
          (int***)RawAllocateWith(MEMMATRIX, sizeof(int**));
          break;

        case 6:
          Vect2_ = Vect3_[0] =
          (int**)RawAllocateWith(MEMMATRIX, sizeof(int*));
          break;

        case 7:
          Vect1_ = Vect2_[0] =
          (int*)RawAllocateWith(MEMMATRIX, sizeof(int) * sz);
          break;

        case 8:
          SetParentReference(Vect1_, Dim_-Off_, x-Off_, Index_-Off_, LenVect_, PosVect_, Error_);
          break;
      }
    }
  }

  return Array_;
}

/****************************************************************************/
// PURPOSE:
//   Method to remove the specified vector from the data matrix given the
//   element index into the matrix where the vector is contained.
//
// PRE:
//   void* Array_     : vector retrieved from the data matrix that is to be
//                      removed.
//   int Dim_         : The dimensions of the matrix, value range = [3...8]
//   int Off_         : The offset into the data matrix, value range = [0...5]
//   int MatrixIndex_ : Index accessing constituent array of the data matrix,
//                      value range = [1...7]
//   int* LenVect_    : Length vector containing length of constituent arrays
//
// POST:
//   The specified vector is removed from the data matrix.
//
void RemoveArray(void* Array_, int Dim_, int Off_, int MatrixIndex_, int* LenVect_)
{
  if (Dim_)
  {
    int y;
    int x = MatrixIndex_;
    int xOff_ = x - Off_;
    int xDim_ = Dim_ - Off_;
    int sz = SizeOfVector(xDim_, xOff_, LenVect_);
    // [i=2...n][(i+1)=3...x+i+1]*

    int******* Vect7_ = (x == 1) ? (int*******)Array_:NULL;
    int****** Vect6_ = (x == 2) ? (int******)Array_:NULL;
    int***** Vect5_ = (x == 3) ? (int*****)Array_:NULL;
    int**** Vect4_ = (x == 4) ? (int****)Array_:NULL;
    int*** Vect3_ = (x == 5) ? (int***)Array_:NULL;
    int** Vect2_ = (x == 6) ? (int**)Array_:NULL;
    int* Vect1_ = (x == 7) ? (int*)Array_:NULL;
    
    int* Contain_ =
    (x == 1) ? Vect7_[0][0][0][0][0][0]:
    (x == 2) ? Vect6_[0][0][0][0][0]:
    (x == 3) ? Vect5_[0][0][0][0]:
    (x == 4) ? Vect4_[0][0][0]:
    (x == 5) ? Vect3_[0][0]:
    (x == 6) ? Vect2_[0]:
    (x == 7) ? Vect1_:NULL;

    // Contain_[pos] = 1;    // deleted
    // Contain_[pos] = 2;    // shared
    // SizeOfVector(Dim_, x, LenVect_)
    // CountOfVector(Dim_, x, Off_, Vect1_, Contain_, Contain2_)
    // StatusOfVector(Dim_, x, Off_, y, Vectl_, Contain_)
    // SetStatusOfVector(Dim_, x, Off_, y, stat, Contain_)

    if (Dim_ <= 2)
    {
      if (Dim_ == 2 && Vect2_)
      {
        for (y = 1; y < sz; y++)
          if (Vect2_[y])
            ::DeleteArray(Vect2_[y]);
      }
      else
        ::DeleteArray(Vect1_);
    }
    else if (1 <= x && x < 7)
    {
      for (y = xOff_+1; y < sz; y++)
      {
        if (x == 1 && Vect7_ && Vect7_[y] &&
            StatusOfVector(Dim_, x, Off_, y, Vect1_, Contain_) == 0)
        {
          SetStatusOfVector(Dim_, x, Off_, y, 1, Contain_);
          RemoveArray(Vect7_[y], Dim_, Off_, x+1, LenVect_);
          ::DeleteArray(Vect7_[y]);
        }
        else if (x == 2 && Vect6_ && Vect6_[y] &&
                 StatusOfVector(Dim_, x, Off_, y, Vect1_, Contain_) == 0)
        {
          SetStatusOfVector(Dim_, x, Off_, y, 1, Contain_);
          RemoveArray(Vect6_[y], Dim_, Off_, x+1, LenVect_);
          ::DeleteArray(Vect6_[y]);
        }
        else if (x == 3 && Vect5_ && Vect5_[y] &&
                 StatusOfVector(Dim_, x, Off_, y, Vect1_, Contain_) == 0)
        {
          SetStatusOfVector(Dim_, x, Off_, y, 1, Contain_);        
          RemoveArray(Vect5_[y], Dim_, Off_, x+1, LenVect_);
          ::DeleteArray(Vect5_[y]);
        }
        else if (x == 4 && Vect4_ && Vect4_[y] &&
                 StatusOfVector(Dim_, x, Off_, y, Vect1_, Contain_) == 0)
        {
          SetStatusOfVector(Dim_, x, Off_, y, 1, Contain_);
          RemoveArray(Vect4_[y], Dim_, Off_, x+1, LenVect_);
          ::DeleteArray(Vect4_[y]);
        }
        else if (x == 5 && Vect3_ && Vect3_[y] &&
                 StatusOfVector(Dim_, x, Off_, y, Vect1_, Contain_) == 0)
        {
          SetStatusOfVector(Dim_, x, Off_, y, 1, Contain_);
          RemoveArray(Vect3_[y], Dim_, Off_, x+1, LenVect_);
          ::DeleteArray(Vect3_[y]);
        }
        else if (x == 6 && Vect2_ && Vect2_[y] &&
                 StatusOfVector(Dim_, x, Off_, y, Vect1_, Contain_) == 0)
        {
          SetStatusOfVector(Dim_, x, Off_, y, 1, Contain_);
          ::DeleteArray(Vect2_[y]);
        }
      }

      if (x == 1 && Vect7_)
        ::DeleteArray(Vect7_);
      else if (x == 2 && Vect6_)
        ::DeleteArray(Vect6_);
      else if (x == 3 && Vect5_)
        ::DeleteArray(Vect5_);
      else if (x == 4 && Vect4_)
        ::DeleteArray(Vect4_);
      else if (x == 5 && Vect3_)
        ::DeleteArray(Vect3_);
      else if (x == 6 && Vect2_)
        ::DeleteArray(Vect2_);      
    }
    else if (x == 7 && Vect1_)
      ::DeleteArray(Vect1_);
  }
}

/****************************************************************************/
// PURPOSE:
//   Method to set ancestor references on an array given an array filled with
//   parent references.
//
// PRE:
//   int* ParentRefs_ : The parent reference array filled with parent
//                      references of the specified vector that is passed.
//   int Dim_         : The dimensions of the matrix, value range = [3...8]
//   int Off_         : The offset into the data matrix, value range = [0...5]
//   int MatrixIndex_ : Index accessing constituent array of the data matrix,
//                      value range = [1...7]
//   int* LenVect_    : Length vector containing length of constituent arrays
//   int* PosVect_    : Position vector containing coordinates into matrix
//   void* Array_     : vector retrieved from the data matrix containing the
//                      array to be assigned parent references.
//   bool& Error_     : Error in array indexing (out of range)
//
// POST:
//   Returns the data matrix with the specified array assigned with parent
//   references.
//
void* SetReferences(int* ParentRefs_, int Dim_, int Off_, int MatrixIndex_,
                    int* LenVect_, int* PosVect_, void* Array_, bool& Error_)
{
  int pos;
  int y;
  int x = MatrixIndex_;
  int xOff_ = x - Off_;
  int xDim_ = Dim_ - Off_;
  int* Vector_;

  int******* Vect7_ = (x == 1) ? (int*******)Array_:NULL;
  int****** Vect6_ = (x == 2) ? (int******)Array_:NULL;
  int***** Vect5_ = (x == 3) ? (int*****)Array_:NULL;
  int**** Vect4_ = (x == 4) ? (int****)Array_:NULL;
  int*** Vect3_ = (x == 5) ? (int***)Array_:NULL;
  int** Vect2_ = (x == 6) ? (int**)Array_:NULL;
  int* Vect1_ = (x == 7) ? (int*)Array_:NULL;

  if (InRange(xOff_+1, PosVect_[xOff_+1], SizeOfVector(xDim_, xOff_, LenVect_) - 1))
    pos = PosVect_[xOff_+1];   // = [(i+1)=3...x+i+1]
  else
    pos = 0;

  Error_ = !pos;  

  if (pos)
  {        
    for (y = 1; y <= ParentRefs_[0]; y++)
    {
      // array[1...n]...[(i+1)=3...x+i+1] = ParentRefs[1...n]
      Vector_ =
      (x == 1) ? Vect7_[y][0][0][0][0][0]:
      (x == 2) ? Vect6_[y][0][0][0][0]:      
      (x == 3) ? Vect5_[y][0][0][0]:
      (x == 4) ? Vect4_[y][0][0]:
      (x == 5) ? Vect3_[y][0]:
      (x == 6) ? Vect2_[y]:
      (x == 7) ? Vect1_:NULL;

      if (Vector_)
        Vector_[pos] = ParentRefs_[y];
    }
  }

  return Array_;
}

/****************************************************************************/
// PURPOSE:
//   Method to remove the parent references from the general data matrix
//   with values from a position vector indexed with a parent index value
//
// array[1...n]
//      [i=2...n][0][0][0][0] = 3rdArr array size=x index=i=2 of
//                              2nd...nth 2ndArr element. 0 = deleted.
//
// array[1...n]
//      [i=2...n][0][0][0][(i+1)=3...x+i+1] = Deallocated array flag = 0/1.
//                                          = Shallow array copy flag = 0/2.
//
// array[1...n]
//      [i=2...n][1][0][0][(i+1)=3...x+i+1] = 3rdArr to TopPar pointer
//      reference number, referencing TopPar pointer in 1st...nth TopArr
//      element for 3rdArr arrays from 3...x.
//
// array[1...n]
//      [i=2...n][i=2][0][0][(i+1)=3...x+i+1] = 3rdArr to 2ndArr pointer
//      reference number, referencing 2ndArr pointer in 2nd...nth 2ndArr
//      element for 3rdArr arrays from 3...x.
//
// array[1...n]
//      [i=2...n][(i+1)=3...x+i+1]* = 1st starting 3rdArr value element to
//                                    ending 3rdArr value element x.
//
// PRE:
//   void* Container_ : Container array 1 level up which contains the
//                      data vector.
//   int Dim_         : The dimensions of the matrix, value range = [3...8]
//   int Off_         : The offset into the data matrix, value range = [0...5]
//   int MatrixIndex_ : Index accessing constituent array of the data matrix,
//                      value range = [1...7]
//   int* LenVect_    : Length vector containing length of constituent arrays
//   int* PosVect_    : Position vector containing coordinates into matrix
//   void* Replacement_ : Replacement vector for deleted existing vector
//                        if not set as NULL, otherwise deleted vector in
//                        data matrix is set as NULL in the array entry.
//   bool DeAlloc_      : Delete the specified array in the data vector rather
//                        than replace it or set the vector size to zero.
//   int* Shared_       : Set to the deallocation status flag of the
//                        container vector of the specified vector to be
//                        deleted. Valid values include: 0=filled,
//                        1=deallocated, 2=shared
//   int* Size_         : Set to the size in elements of the container vector.
//   int* Max_          : Set to the maximum length of the container vector.
//   bool& Error_       : Error in array indexing (out of range)
//
void* RemoveReferences(void* Container_, int Dim_, int Off_,
                       int MatrixIndex_, int* LenVect_, int* PosVect_,
                       void* Replacement_, bool DeAlloc_, int* Shared_,
                       int* Size_, int* Max_, bool& Error_)
{
  int CurPos_;
  int PrevPos_;
  int x = MatrixIndex_;
  int lenpos = Dim_ - x;
  int xOff_ = x - Off_;
  int xDim_ = Dim_ - Off_;
  bool inc;

  if (InRange(xOff_, PosVect_[xOff_], SizeOfVector(xDim_, xOff_-1, LenVect_) - 1))
    PrevPos_ = PosVect_[xOff_];
  else
    PrevPos_ = 0;

  if (x < 7)
  {
    if (InRange(xOff_+1, PosVect_[xOff_+1], SizeOfVector(xDim_, xOff_, LenVect_) - 1))
      CurPos_ = PosVect_[xOff_+1];
    else
      CurPos_ = 0;
  }
  else
    CurPos_ = 0;

  Error_ = PrevPos_ == 0 || (x < 7 && CurPos_ == 0);
  
  if (Error_)
    return NULL;    

  int******** Vect8_ = (x==1) ? (int********)Container_:NULL;
  int******* Vect7_ = (x==2) ? (int*******)Container_:((Vect8_ && x==1) ? Vect8_[PrevPos_]:NULL);
  int****** Vect6_ = (x==3) ? (int******)Container_:((Vect7_ && x==2) ? Vect7_[PrevPos_]:NULL);
  int***** Vect5_ = (x==4) ? (int*****)Container_:((Vect6_ && x==3) ? Vect6_[PrevPos_]:NULL);
  int**** Vect4_ = (x==5) ? (int****)Container_:((Vect5_ && x==4) ? Vect5_[PrevPos_]:NULL);
  int*** Vect3_ = (x==6) ? (int***)Container_:((Vect4_ && x==5) ? Vect4_[PrevPos_]:NULL);
  int** Vect2_ = (x==7) ? (int**)Container_:((Vect3_ && x==6) ? Vect3_[PrevPos_]:NULL);
  int* Vect1_ = (x==7) ? (Vect2_ ? Vect2_[PrevPos_]:NULL):NULL;

  int pos = PrevPos_;
  int* Vector_;
  int** Contain2_;
  int* Contain_;
  
  int* ParentRefs_ =
  (x < 7) ? ((int*)RawAllocateWith(MEMMATRIX, sizeof(int) * (xDim_ + 2))):NULL;
  ParentRefs_[0] = ParentRefs_[xDim_ + 1] = 0;

  if (x < 7)
  {      
    Vector_ =
    (x == 1 && Vect7_) ? Vect7_[0][0][0][0][0][0]:
    (x == 2 && Vect6_) ? Vect6_[0][0][0][0][0]:
    (x == 3 && Vect5_) ? Vect5_[0][0][0][0]:
    (x == 4 && Vect4_) ? Vect4_[0][0][0]:
    (x == 5 && Vect3_) ? Vect3_[0][0]:
    (x == 6 && Vect2_) ? Vect2_[0]:NULL;

    Contain2_ =
    (x == 1 && Vect8_) ? Vect8_[0][0][0][0][0][0]:
    (x == 2 && Vect7_) ? Vect7_[0][0][0][0][0]:
    (x == 3 && Vect6_) ? Vect6_[0][0][0][0]:
    (x == 4 && Vect5_) ? Vect5_[0][0][0]:
    (x == 5 && Vect4_) ? Vect4_[0][0]:
    (x == 6 && Vect3_) ? Vect3_[0]:NULL;

    Contain_ = Contain2_ ? Contain2_[0]:NULL;
  }
  else
  {
    Vector_ = Vect1_;
    Contain2_ = Vect2_;
    Contain_ = Vect2_[0];
  }

  // Contain_[pos] = 1;    // deleted
  // Contain_[pos] = 2;    // shared
  // SizeOfVector(Dim_, x, LenVect_)
  // CountOfVector(Dim_, x, Off_, Vect1_, Contain_, Contain2_)
  // StatusOfVector(Dim_, x, Off_, y, Vectl_, Contain_)
  // SetStatusOfVector(Dim_, x, Off_, y, stat, Contain_)
  
  // 0th index of parent reference array is the size of the array
  switch (x)
  {
    case 1:
      if (Vect8_[pos] || Replacement_)
      {
        if (Replacement_)
        {
          inc = !Vect7_ || !Vector_[0];
          if (inc)                            // increment size
          {
            IncCountOfVector(Dim_, x, Off_, Vect1_, Contain_, Contain2_);
            if (CountOfVector(Dim_, x, Off_, Vect1_, Contain_, Contain2_) > Contain_[0])
              ++Contain_[0];
          }

          Vect8_[pos] = (int*******)Replacement_;
          if (Vect7_ && Vect7_[xOff_] && Vect8_[pos][xOff_])
            GetParentReference(ParentRefs_, Dim_, Off_, MatrixIndex_,
                               LenVect_, PosVect_, Vect7_, Error_);

          RemoveArray(Vect7_, Dim_, Off_, MatrixIndex_, LenVect_);
          Vect7_ = Vect8_[pos];
          SetStatusOfVector(Dim_, x, Off_, pos, 2, Contain_);
          Vector_ = Vect7_[0][0][0][0][0][0];

          if (inc)
            Vector_[0] = LenVect_[lenpos];    // array[n]...[0] = Length[n]

          if (Vect7_[xOff_])
            SetReferences(ParentRefs_, Dim_, Off_, MatrixIndex_,
                          LenVect_, PosVect_, Vect7_, Error_);
          else
            SetRefArrays(Vect7_, Dim_, Off_, MatrixIndex_, LenVect_, PosVect_, Error_);
        }
        else
        {
          // Setting vector size
          if (CountOfVector(Dim_, x, Off_, Vect1_, Contain_, Contain2_) > 0)       // decrement size
            DecCountOfVector(Dim_, x, Off_, Vect1_, Contain_, Contain2_);

          if (DeAlloc_)
          {
            RemoveArray(Vect7_, Dim_, Off_, MatrixIndex_, LenVect_);
            Vect7_ = Vect8_[pos] = NULL;
            SetStatusOfVector(Dim_, x, Off_, pos, 1, Contain_);
            Vector_ = NULL;
          }
          else
            Vector_[0] = 0;          // subarray size = 0, (deleted)
        }

        if (Shared_)
          *Shared_ = StatusOfVector(Dim_, x, Off_, pos, Vect1_, Contain_);
          
        if (Size_)
          *Size_ = CountOfVector(Dim_, x, Off_, Vect1_, Contain_, Contain2_);

        if (Max_)
          *Max_ = Contain_[0];
      }
      break;
    
    case 2:
      if (Vect7_[pos] || Replacement_)
      {
        if (Replacement_)
        {
          inc = !Vect6_ || !Vector_[0];
          if (inc)                            // increment size
          {
            IncCountOfVector(Dim_, x, Off_, Vect1_, Contain_, Contain2_);
            if (CountOfVector(Dim_, x, Off_, Vect1_, Contain_, Contain2_) > Contain_[0])
              ++Contain_[0];
          }

          Vect7_[pos] = (int******)Replacement_;
          if (Vect6_ && Vect6_[xOff_] && Vect7_[pos][xOff_])
            GetParentReference(ParentRefs_, Dim_, Off_, MatrixIndex_,
                               LenVect_, PosVect_, Vect6_, Error_);

          RemoveArray(Vect6_, Dim_, Off_, MatrixIndex_, LenVect_);
          Vect6_ = Vect7_[pos];
          SetStatusOfVector(Dim_, x, Off_, pos, 2, Contain_);
          Vector_ = Vect6_[0][0][0][0][0];

          if (inc)
            Vector_[0] = LenVect_[lenpos];    // array[n]...[0] = Length[n]

          if (Vect6_[xOff_])
            SetReferences(ParentRefs_, Dim_, Off_, MatrixIndex_,
                          LenVect_, PosVect_, Vect6_, Error_);
          else
            SetRefArrays(Vect6_, Dim_, Off_, MatrixIndex_, LenVect_, PosVect_, Error_);
        }
        else
        {
          // Setting vector size
          if (CountOfVector(Dim_, x, Off_, Vect1_, Contain_, Contain2_) > 0)       // decrement size
            DecCountOfVector(Dim_, x, Off_, Vect1_, Contain_, Contain2_);

          if (DeAlloc_)
          {
            RemoveArray(Vect6_, Dim_, Off_, MatrixIndex_, LenVect_);
            Vect6_ = Vect7_[pos] = NULL;
            SetStatusOfVector(Dim_, x, Off_, pos, 1, Contain_);
            Vector_ = NULL;
          }
          else
            Vector_[0] = 0;          // subarray size = 0, (deleted)
        }

        if (Shared_)
          *Shared_ = StatusOfVector(Dim_, x, Off_, pos, Vect1_, Contain_);

        if (Size_)
          *Size_ = CountOfVector(Dim_, x, Off_, Vect1_, Contain_, Contain2_);

        if (Max_)
          *Max_ = Contain_[0];        
      }
      break;
    
    case 3:
      if (Vect6_[pos] || Replacement_)
      {
        if (Replacement_)
        {
          inc = !Vect5_ || !Vector_[0];
          if (inc)                            // increment size
          {
            IncCountOfVector(Dim_, x, Off_, Vect1_, Contain_, Contain2_);
            if (CountOfVector(Dim_, x, Off_, Vect1_, Contain_, Contain2_) > Contain_[0])
              ++Contain_[0];
          }          

          Vect6_[pos] = (int*****)Replacement_;
          if (Vect5_ && Vect5_[xOff_] && Vect6_[pos][xOff_])
            GetParentReference(ParentRefs_, Dim_, Off_, MatrixIndex_,
                               LenVect_, PosVect_, Vect5_, Error_);

          RemoveArray(Vect5_, Dim_, Off_, MatrixIndex_, LenVect_);
          Vect5_ = Vect6_[pos];
          SetStatusOfVector(Dim_, x, Off_, pos, 2, Contain_);          
          Vector_ = Vect5_[0][0][0][0];

          if (inc)
            Vector_[0] = LenVect_[lenpos];    // array[n]...[0] = Length[n]

          if (Vect5_[xOff_])
            SetReferences(ParentRefs_, Dim_, Off_, MatrixIndex_,
                          LenVect_, PosVect_, Vect5_, Error_);
          else
            SetRefArrays(Vect5_, Dim_, Off_, MatrixIndex_, LenVect_, PosVect_, Error_);
        }
        else
        {
          // Setting vector size
          if (CountOfVector(Dim_, x, Off_, Vect1_, Contain_, Contain2_) > 0)       // decrement size
            DecCountOfVector(Dim_, x, Off_, Vect1_, Contain_, Contain2_);

          if (DeAlloc_)
          {
            RemoveArray(Vect5_, Dim_, Off_, MatrixIndex_, LenVect_);
            Vect5_ = Vect6_[pos] = NULL;
            SetStatusOfVector(Dim_, x, Off_, pos, 1, Contain_);            
            Vector_ = NULL;
          }
          else
            Vector_[0] = 0;          // subarray size = 0, (deleted)
        }

        if (Shared_)
          *Shared_ = StatusOfVector(Dim_, x, Off_, pos, Vect1_, Contain_);

        if (Size_)
          *Size_ = CountOfVector(Dim_, x, Off_, Vect1_, Contain_, Contain2_);

        if (Max_)
          *Max_ = Contain_[0];        
      }      
      break;

    case 4:
      if (Vect5_[pos] || Replacement_)
      {
        if (Replacement_)
        {
          inc = !Vect4_ || !Vector_[0];
          if (inc)                            // increment size
          {
            IncCountOfVector(Dim_, x, Off_, Vect1_, Contain_, Contain2_);
            if (CountOfVector(Dim_, x, Off_, Vect1_, Contain_, Contain2_) > Contain_[0])
              ++Contain_[0];          
          }          

          Vect5_[pos] = (int****)Replacement_;
          if (Vect4_ && Vect4_[xOff_] && Vect5_[pos][xOff_])
            GetParentReference(ParentRefs_, Dim_, Off_, MatrixIndex_,
                               LenVect_, PosVect_, Vect4_, Error_);

          RemoveArray(Vect4_, Dim_, Off_, MatrixIndex_, LenVect_);
          Vect4_ = Vect5_[pos];
          SetStatusOfVector(Dim_, x, Off_, pos, 2, Contain_);          
          Vector_ = Vect4_[0][0][0];

          if (inc)
            Vector_[0] = LenVect_[lenpos];    // array[n]...[0] = Length[n]

          if (Vect4_[xOff_])
            SetReferences(ParentRefs_, Dim_, Off_, MatrixIndex_,
                          LenVect_, PosVect_, Vect4_, Error_);
          else
            SetRefArrays(Vect4_, Dim_, Off_, MatrixIndex_, LenVect_, PosVect_, Error_);
        }
        else
        {
          // Setting vector size
          if (CountOfVector(Dim_, x, Off_, Vect1_, Contain_, Contain2_) > 0)       // decrement size
            DecCountOfVector(Dim_, x, Off_, Vect1_, Contain_, Contain2_);

          if (DeAlloc_)
          {
            RemoveArray(Vect4_, Dim_, Off_, MatrixIndex_, LenVect_);
            Vect4_ = Vect5_[pos] = NULL;
            SetStatusOfVector(Dim_, x, Off_, pos, 1, Contain_);
            Vector_ = NULL;
          }
          else
            Vector_[0] = 0;          // subarray size = 0, (deleted)
        }

        if (Shared_)
          *Shared_ = StatusOfVector(Dim_, x, Off_, pos, Vect1_, Contain_);

        if (Size_)
          *Size_ = CountOfVector(Dim_, x, Off_, Vect1_, Contain_, Contain2_);

        if (Max_)
          *Max_ = Contain_[0];
      }      
      break;
    
    case 5:
      if (Vect4_[pos] || Replacement_)
      {
        if (Replacement_)
        {
          inc = !Vect3_ || !Vector_[0];
          if (inc)                            // increment size
          {
            IncCountOfVector(Dim_, x, Off_, Vect1_, Contain_, Contain2_);
            if (CountOfVector(Dim_, x, Off_, Vect1_, Contain_, Contain2_) > Contain_[0])
              ++Contain_[0];          
          }

          Vect4_[pos] = (int***)Replacement_;
          if (Vect3_ && Vect3_[xOff_] && Vect4_[pos][xOff_])
            GetParentReference(ParentRefs_, Dim_, Off_, MatrixIndex_,
                               LenVect_, PosVect_, Vect3_, Error_);

          RemoveArray(Vect3_, Dim_, Off_, MatrixIndex_, LenVect_);
          Vect3_ = Vect4_[pos];
          SetStatusOfVector(Dim_, x, Off_, pos, 2, Contain_);
          Vector_ = Vect3_[0][0];

          if (inc)
            Vector_[0] = LenVect_[lenpos];    // array[n]...[0] = Length[n]

          if (Vect3_[xOff_])
            SetReferences(ParentRefs_, Dim_, Off_, MatrixIndex_,
                          LenVect_, PosVect_, Vect3_, Error_);
          else
            SetRefArrays(Vect3_, Dim_, Off_, MatrixIndex_, LenVect_, PosVect_, Error_);
        }
        else
        {
          // Setting vector size
          if (CountOfVector(Dim_, x, Off_, Vect1_, Contain_, Contain2_) > 0)       // decrement size
            DecCountOfVector(Dim_, x, Off_, Vect1_, Contain_, Contain2_);
          
          if (DeAlloc_)
          {
            RemoveArray(Vect3_, Dim_, Off_, MatrixIndex_, LenVect_);
            Vect3_ = Vect4_[pos] = NULL;
            SetStatusOfVector(Dim_, x, Off_, pos, 1, Contain_);            
            Vector_ = NULL;
          }
          else
            Vector_[0] = 0;          // subarray size = 0, (deleted)
        }

        if (Shared_)
          *Shared_ = StatusOfVector(Dim_, x, Off_, pos, Vect1_, Contain_);

        if (Size_)
          *Size_ = CountOfVector(Dim_, x, Off_, Vect1_, Contain_, Contain2_);

        if (Max_)
          *Max_ = Contain_[0];
      }      
      break;
    
    case 6:
      if (Vect3_[pos] || Replacement_)
      {
        if (Replacement_)
        {
          inc = !Vect2_ || !Vector_[0];
          if (inc)                            // increment size
          {
            IncCountOfVector(Dim_, x, Off_, Vect1_, Contain_, Contain2_);
            if (CountOfVector(Dim_, x, Off_, Vect1_, Contain_, Contain2_) > Contain_[0])
              ++Contain_[0];
          }

          Vect3_[pos] = (int**)Replacement_;
          if (Vect2_ && Vect2_[xOff_] && Vect3_[pos][xOff_])
            GetParentReference(ParentRefs_, Dim_, Off_, MatrixIndex_,
                               LenVect_, PosVect_, Vect2_, Error_);

          RemoveArray(Vect2_, Dim_, Off_, MatrixIndex_, LenVect_);
          Vect2_ = Vect3_[pos];
          SetStatusOfVector(Dim_, x, Off_, pos, 2, Contain_);
          Vector_ = Vect2_[0];

          if (inc)
            Vector_[0] = LenVect_[lenpos];    // array[n]...[0] = Length[n]

          if (Vect2_[xOff_])
            SetReferences(ParentRefs_, Dim_, Off_, MatrixIndex_,
                          LenVect_, PosVect_, Vect2_, Error_);
          else
            SetRefArrays(Vect2_, Dim_, Off_, MatrixIndex_, LenVect_, PosVect_, Error_);
        }
        else
        {
          // Setting vector size
          if (CountOfVector(Dim_, x, Off_, Vect1_, Contain_, Contain2_) > 0)       // decrement size
            DecCountOfVector(Dim_, x, Off_, Vect1_, Contain_, Contain2_);
          
          if (DeAlloc_)
          {
            RemoveArray(Vect2_, Dim_, Off_, MatrixIndex_, LenVect_);
            Vect2_ = Vect3_[pos] = NULL;
            SetStatusOfVector(Dim_, x, Off_, pos, 1, Contain_);            
            Vector_ = NULL;
          }
          else
            Vector_[0] = 0;          // subarray size = 0, (deleted)
        }

        if (Shared_)
          *Shared_ = StatusOfVector(Dim_, x, Off_, pos, Vect1_, Contain_);

        if (Size_)
          *Size_ = CountOfVector(Dim_, x, Off_, Vect1_, Contain_, Contain2_);

        if (Max_)
          *Max_ = Contain_[0];
      }      
      break;
    
    case 7:
      if (Vect2_[pos] || Replacement_)
      {      
        if (Replacement_)
        {
          inc = !Vect1_ || !Vector_[0];
          if (inc)                    // increment size
          {
            IncCountOfVector(Dim_, x, Off_, Vect1_, Contain_, Contain2_);
            if (CountOfVector(Dim_, x, Off_, Vect1_, Contain_, Contain2_) > Contain_[0])
              ++Contain_[0];
          }
          
          ::DeleteArray(Vect1_);
          Vect1_ = Vect2_[pos] = (int*)Replacement_;
          SetStatusOfVector(Dim_, x, Off_, pos, 2, Contain_);
          Vector_ = Vect1_;

          if (inc)
            Vector_[0] = LenVect_[lenpos];  // array[n]...[0] = Length[n]
        }
        else
        {
          // Setting vector size
          if (CountOfVector(Dim_, x, Off_, Vect1_, Contain_, Contain2_) > 0)       // decrement size
            DecCountOfVector(Dim_, x, Off_, Vect1_, Contain_, Contain2_);

          if (DeAlloc_)
          {
            ::DeleteArray(Vect1_);
            Vect1_ = Vect2_[pos] = NULL;
            SetStatusOfVector(Dim_, x, Off_, pos, 1, Contain_);            
            Vector_ = NULL;
          }
          else
            Vector_[0] = 0;          // subarray size = 0, (deleted)
        }

        if (Shared_)
          *Shared_ = StatusOfVector(Dim_, x, Off_, pos, Vect1_, Contain_);

        if (Size_)
          *Size_ = CountOfVector(Dim_, x, Off_, Vect1_, Contain_, Contain2_);

        if (Max_)
          *Max_ = Contain_[0];
      }
  }

  ::DeleteArray(ParentRefs_);
  return Container_;
}

/****************************************************************************/
// PURPOSE:
//   Method to set the reference array for each constituent array
//
// array[1...n]
//      [i=2...n][0][0][0][0] = 3rdArr array size=x index=i=2 of
//                              2nd...nth 2ndArr element. 0 = deleted.
//
// array[1...n]
//      [i=2...n][0][0][0][(i+1)=3...x+i+1] = Deallocated array flag = 0/1.
//                                          = Shallow array copy flag = 0/2.
//
// array[1...n]
//      [i=2...n][1][0][0][(i+1)=3...x+i+1] = 3rdArr to TopPar pointer
//      reference number, referencing TopPar pointer in 1st...nth TopArr
//      element for 3rdArr arrays from 3...x.
//
// array[1...n]
//      [i=2...n][i=2][0][0][(i+1)=3...x+i+1] = 3rdArr to 2ndArr pointer
//      reference number, referencing 2ndArr pointer in 2nd...nth 2ndArr
//      element for 3rdArr arrays from 3...x.
//
// array[1...n]
//      [i=2...n][(i+1)=3...x+i+1]* = 1st starting 3rdArr value element to
//                                    ending 3rdArr value element x.
//
// PRE:
//   void* Array   : Data matrix
//   int Dim_      : The dimensions of the matrix, value range = [2...8]
//   int Index_    : Index level into position vector, denoted by constant
//                   i == Index_, max = Index_ + 1
//   int* LenVect_ : Length vector containing length of constituent arrays
//   int* PosVect_ : Position vector containing coordinates into matrix
//
//   void* Replacement_ : Replacement vector for deleted existing vector
//                        if not set as NULL, otherwise deleted vector in
//                        data matrix is set as NULL in the array entry.
//                        Shared_ set to TRUE if replacement vector present,
//                        otherwise Shared_ is set to FALSE.
//   bool Replace_      : If set to true then if Replacement_ is NULL a new
//                        vector is created and then that is used as the
//                        replacement vector, otherwise if false then if
//                        Replacement_ is NULL the element position in the
//                        data matrix is deleted and set to NULL.
//   bool DeAlloc_      : Should actually perform memory deallocation in
//                        data matrix element position instead of setting
//                        vector size to zero.
//   bool& Error_       : Error in array indexing (out of range)
//   bool CheckDealloc_ : Do not perform deletion or replacement, but
//                        retrieve size and deallocation/shared information
//                        of the specified vector from the data matrix.
//   int* Shared_       : Set to the deallocation status flag of the
//                        container vector of the specified vector to be
//                        deleted. Valid values include: 0=filled,
//                        1=deallocated, 2=shared
//   int* Size_         : Set to the size in elements of the container vector.
//   int* Max_          : Set to the maximum length of the container vector.
//
// POST:
//   Returns data matrix with specified constituent vector deleted and/or
//   replaced and has containing array entry set to NULL or has containing
//   array entry shared with replacement vector. If CheckDealloc is set TRUE
//   then no deletion or replacement is actually performed, but size and
//   deallocation/shared information of the specified vector from the data
//   matrix is retrieved.
//
int DeleteVectorHelper(void* Array_, int Dim_, int Index_, int* LenVect_,
                       int* PosVect_, void* Replacement_, bool Replace_,
                       bool DeAlloc_, bool& Error_, bool CheckDealloc_,
                       int* Shared_, int* Size_, int* Max_)
{
  int******** Vect8_;
  int******* Vect7_;
  int****** Vect6_;
  int***** Vect5_;
  int**** Vect4_;
  int*** Vect3_;
  int** Vect2_;
  int* Vect1_;
  int** Contain2_;
  int* Contain_;

  int off = 8 - Dim_; // Maximum dimensions is: 8
  int pos;  
  int inc;
  int sz;
  int x;

  int xIndex_;
  int xDim_;
  int xOff_;

  // Contain_[pos] = 1;    // deleted
  // Contain_[pos] = 2;    // shared
  // SizeOfVector(Dim_, x, LenVect_)
  // CountOfVector(Dim_, x, Off_, Vect1_, Contain_, Contain2_)
  // StatusOfVector(Dim_, x, Off_, y, Vect1_, Contain_)
  // SetStatusOfVector(Dim_, x, Off_, y, stat, Contain_)  

  if (2 <= Dim_ && Dim_ <= 8 &&
      0 <= Index_ && Index_ < Dim_)
  {
    if (Index_ != 0)
    {
      if (Dim_ == 2)
      {
        if (InRange(1, PosVect_[1], SizeOfVector(Dim_, 0, LenVect_) - 1))
          pos = PosVect_[1];
        else
          pos = 0;

        if (pos)
        {
          Vect2_ = (int**)Array_;

          if (CheckDealloc_ && Vect2_ && Vect2_[0])
          {
            if (Shared_)
              *Shared_ = Vect2_[0][pos];

            if (Max_)
              *Max_ = Vect2_[0][0];

            return (Shared_ || Size_);
          }
          else if (Vect2_[pos] || Replacement_ || Replace_)
          {
            Vect1_ = Vect2_[pos];

            if (Replacement_ || Replace_)
            {
              inc = !Vect1_ || !Vect1_[0];
              
              ::DeleteArray(Vect1_);
              if (Replace_ && !Replacement_)
              {
                sz = SizeOfVector(Dim_, 1, LenVect_);
                Vect1_ = (int*)RawAllocateWith(MEMMATRIX, sizeof(int) * sz);
                memset(Vect1_, 0, sizeof(int) * sz);
              }
              
              Vect1_ = Vect2_[pos] = (int*)Replacement_;
              Vect2_[0][pos] = 2;    // shared

              if (inc)
                Vect1_[0] = LenVect_[1];  // array[n]...[0] = Length[n]
            }
            else
            {
              if (DeAlloc_)
              {
                ::DeleteArray(Vect1_);
                Vect1_ = Vect2_[pos] = NULL;
                Vect2_[0][pos] = 1;  // deleted
              }
              else
                Vect1_[0] = 0;            // subarray size = 0, (deleted)
            }

            if (Shared_)
              *Shared_ = Vect2_[0][pos];

            if (Max_)
              *Max_ = Vect2_[0][0];            
          }

          Error_ = false;
        }
        else
        {
          Error_ = true;
          return NULL;
        }
      }
      else
      {
        Vect8_ = (off == 0) ? (int********)Array_:NULL;
        Vect7_ = (off == 1) ? (int*******)Array_:NULL;
        Vect6_ = (off == 2) ? (int******)Array_:NULL;
        Vect5_ = (off == 3) ? (int*****)Array_:NULL;
        Vect4_ = (off == 4) ? (int****)Array_:NULL;
        Vect3_ = (off == 5) ? (int***)Array_:NULL;
        Vect2_ = (off == 6) ? (int**)Array_:NULL;

        xDim_ = Dim_;
        Dim_ += off;
        xIndex_ = Index_;
        Index_ += off;

        for (x = 1; x < xDim_; x++)
        {
          xOff_ = x + off;          

          if ((Error_ = (x <= xIndex_)) &&
              InRange(x, PosVect_[x], SizeOfVector(xDim_, x-1, LenVect_) - 1))
            pos = PosVect_[x];
          else
            pos = 0;

          if (!pos && Error_)
            return NULL;
          else
            Error_ = false;
                        
          switch (xOff_)
          {
            case 1:
              if (pos)
              {
                if (xOff_ == Index_)
                {
                  if (CheckDealloc_ && Vect8_ && Vect8_[0])
                  {
                    Contain2_ = Vect8_[0][0][0][0][0][0];
                    Contain_ = Contain2_[0];
                    
                    if (Shared_)
                      *Shared_ = StatusOfVector(Dim_, Index_, off, pos, Vect1_, Contain_);
                      
                    if (Size_)
                      *Size_ = CountOfVector(Dim_, Index_, off, Vect1_, Contain_, Contain2_);

                    if (Max_)
                      *Max_ = Contain_[0];

                    return (Shared_ || Size_);
                  }
                  else
                  {
                    if (Replace_ && !Replacement_)
                    {
                      sz = SizeOfVector(xDim_, xIndex_, LenVect_);
                      Vect7_ = (int*******)RawAllocateWith(MEMMATRIX, sizeof(int******) * sz);
                      Replacement_ = InitMatrix(Vect7_, Dim_, off, Index_, LenVect_);
                      Replacement_ = SetRefArrays(Vect7_, Dim_, off, Index_, LenVect_, PosVect_, Error_);
                    }
                  
                    if (Vect8_[pos])
                    {
                      Vect7_ = Vect8_[pos];
                      RemoveReferences(Vect8_, Dim_, off, Index_, LenVect_,
                                       PosVect_, Replacement_, DeAlloc_,
                                       Shared_, Size_, Max_, Error_);
                    }
                    else if (Replacement_)
                    {
                      Vect8_[pos] = Vect7_ = (int*******)Replacement_;
                      Contain2_ = Vect8_[0][0][0][0][0][0];
                      Contain_ = Contain2_[0];

                      IncCountOfVector(Dim_, Index_, off, Vect1_, Contain_, Contain2_);
                      if (CountOfVector(Dim_, Index_, off, Vect1_, Contain_, Contain2_) > Contain_[0])
                        ++Contain_[0];
                    }
                  }
                }                
                else
                {
                  if (Vect8_[pos])
                    Vect7_ = Vect8_[pos];
                  else
                    return 0;
                }
              }
              break;

            case 2:
              if (pos)
              {
                if (xOff_ == Index_)
                {
                  if (CheckDealloc_ && Vect7_ && Vect7_[0])
                  {
                    Contain2_ = Vect7_[0][0][0][0][0];
                    Contain_ = Contain2_[0];
                    
                    if (Shared_)
                      *Shared_ = StatusOfVector(Dim_, Index_, off, pos, Vect1_, Contain_);

                    if (Size_)
                      *Size_ = CountOfVector(Dim_, Index_, off, Vect1_, Contain_, Contain2_);

                    if (Max_)
                      *Max_ = Contain_[0];

                    return (Shared_ || Size_);
                  }
                  else
                  {
                    if (Replace_ && !Replacement_)
                    {
                      sz = SizeOfVector(xDim_, xIndex_, LenVect_);
                      Vect6_ = (int******)RawAllocateWith(MEMMATRIX, sizeof(int*****) * sz);
                      Replacement_ = InitMatrix(Vect6_, Dim_, off, Index_, LenVect_);
                      Replacement_ = SetRefArrays(Vect6_, Dim_, off, Index_, LenVect_, PosVect_, Error_);
                    }
                  
                    if (Vect7_[pos])                    
                    {
                      Vect6_ = Vect7_[pos];
                      RemoveReferences(Vect7_, Dim_, off, Index_, LenVect_,
                                       PosVect_, Replacement_, DeAlloc_,
                                       Shared_, Size_, Max_, Error_);
                    }
                    else if (Replacement_)
                    {                    
                      Vect7_[pos] = Vect6_ = (int******)Replacement_;
                      Contain2_ = Vect7_[0][0][0][0][0];
                      Contain_ = Contain2_[0];
                      
                      IncCountOfVector(Dim_, Index_, off, Vect1_, Contain_, Contain2_);
                      if (CountOfVector(Dim_, Index_, off, Vect1_, Contain_, Contain2_) > Contain_[0])
                        ++Contain_[0];
                    }
                  }
                }
                else
                {
                  if (Vect7_[pos])
                    Vect6_ = Vect7_[pos];
                  else
                    return 0;
                }
              }
              else if (x != 1)
              {
                if (Vect7_[0])
                  Vect6_ = Vect7_[0];
                else
                  return 0;
              }              
              break;
              
            case 3:
              if (pos)
              {
                if (xOff_ == Index_)
                {
                  if (CheckDealloc_ && Vect6_ && Vect6_[0])
                  {
                    Contain2_ = Vect6_[0][0][0][0];
                    Contain_ = Contain2_[0];

                    if (Shared_)
                      *Shared_ = StatusOfVector(Dim_, Index_, off, pos, Vect1_, Contain_);
                      
                    if (Size_)
                      *Size_ = CountOfVector(Dim_, Index_, off, Vect1_, Contain_, Contain2_);

                    if (Max_)
                      *Max_ = Contain_[0];

                    return (Shared_ || Size_);
                  }
                  else
                  {
                    if (Replace_ && !Replacement_)
                    {
                      sz = SizeOfVector(xDim_, xIndex_, LenVect_);
                      Vect5_ = (int*****)RawAllocateWith(MEMMATRIX, sizeof(int****) * sz);
                      Replacement_ = InitMatrix(Vect5_, Dim_, off, Index_, LenVect_);
                      Replacement_ = SetRefArrays(Vect5_, Dim_, off, Index_, LenVect_, PosVect_, Error_);
                    }
                  
                    if (Vect6_[pos])
                    {
                      Vect5_ = Vect6_[pos];
                      RemoveReferences(Vect6_, Dim_, off, Index_, LenVect_,
                                       PosVect_, Replacement_, DeAlloc_,
                                       Shared_, Size_, Max_, Error_);
                    }
                    else if (Replacement_)
                    {
                      Vect6_[pos] = Vect5_ = (int*****)Replacement_;
                      Contain2_ = Vect6_[0][0][0][0];
                      Contain_ = Contain2_[0];

                      IncCountOfVector(Dim_, Index_, off, Vect1_, Contain_, Contain2_);
                      if (CountOfVector(Dim_, Index_, off, Vect1_, Contain_, Contain2_) > Contain_[0])
                        ++Contain_[0];
                    }
                  }
                }
                else
                {
                  if (Vect6_[pos])
                    Vect5_ = Vect6_[pos];
                  else
                    return 0;
                }
              }
              else if (x != 1)
              {
                if (Vect6_[0])
                  Vect5_ = Vect6_[0];
                else
                  return 0;
              }              
              break;
            
            case 4:
              if (pos)
              {
                if (xOff_ == Index_)
                {
                  if (CheckDealloc_ && Vect5_ && Vect5_[0])
                  {
                    Contain2_ = Vect5_[0][0][0];
                    Contain_ = Contain2_[0];

                    if (Shared_)
                      *Shared_ = StatusOfVector(Dim_, Index_, off, pos, Vect1_, Contain_);
                      
                    if (Size_)
                      *Size_ = CountOfVector(Dim_, Index_, off, Vect1_, Contain_, Contain2_);

                    if (Max_)
                      *Max_ = Contain_[0];
                  
                    return (Shared_ || Size_);
                  }
                  else
                  {
                    if (Replace_ && !Replacement_)
                    {
                      sz = SizeOfVector(xDim_, xIndex_, LenVect_);
                      Vect4_ = (int****)RawAllocateWith(MEMMATRIX, sizeof(int***) * sz);
                      Replacement_ = InitMatrix(Vect4_, Dim_, off, Index_, LenVect_);
                      Replacement_ = SetRefArrays(Vect4_, Dim_, off, Index_, LenVect_, PosVect_, Error_);
                    }
                  
                    if (Vect5_[pos])
                    {
                      Vect4_ = Vect5_[pos];
                      RemoveReferences(Vect5_, Dim_, off, Index_, LenVect_,
                                       PosVect_, Replacement_, DeAlloc_,
                                       Shared_, Size_, Max_, Error_);
                    }
                    else if (Replacement_)
                    {
                      Vect5_[pos] = Vect4_ = (int****)Replacement_;
                      Contain2_ = Vect5_[0][0][0];
                      Contain_ = Contain2_[0];

                      IncCountOfVector(Dim_, Index_, off, Vect1_, Contain_, Contain2_);
                      if (CountOfVector(Dim_, Index_, off, Vect1_, Contain_, Contain2_) > Contain_[0])
                        ++Contain_[0];
                    }
                  }
                }
                else
                {
                  if (Vect5_[pos])
                    Vect4_ = Vect5_[pos];
                  else
                    return 0;
                }
              }
              else if (x != 1)
              {
                if (Vect5_[0])
                  Vect4_ = Vect5_[0];
                else
                  return 0;
              }
              break;
            
            case 5:
              if (pos)
              {
                if (xOff_ == Index_)
                {
                  if (CheckDealloc_ && Vect4_ && Vect4_[0])
                  {
                    Contain2_ = Vect4_[0][0];
                    Contain_ = Contain2_[0];

                    if (Shared_)
                      *Shared_ = StatusOfVector(Dim_, Index_, off, pos, Vect1_, Contain_);
                      
                    if (Size_)
                      *Size_ = CountOfVector(Dim_, Index_, off, Vect1_, Contain_, Contain2_);

                    if (Max_)
                      *Max_ = Contain_[0];
                  
                    return (Shared_ || Size_);
                  }
                  else
                  {
                    if (Replace_ && !Replacement_)
                    {
                      sz = SizeOfVector(xDim_, xIndex_, LenVect_);
                      Vect3_ = (int***)RawAllocateWith(MEMMATRIX, sizeof(int**) * sz);
                      Replacement_ = InitMatrix(Vect3_, Dim_, off, Index_, LenVect_);
                      Replacement_ = SetRefArrays(Vect3_, Dim_, off, Index_, LenVect_, PosVect_, Error_);
                    }
                  
                    if (Vect4_[pos])
                    {
                      Vect3_ = Vect4_[pos];
                      RemoveReferences(Vect4_, Dim_, off, Index_, LenVect_,
                                       PosVect_, Replacement_, DeAlloc_,
                                       Shared_, Size_, Max_, Error_);
                    }
                    else if (Replacement_)
                    {
                      Vect4_[pos] = Vect3_ = (int***)Replacement_;
                      Contain2_ = Vect4_[0][0];
                      Contain_ = Contain2_[0];

                      IncCountOfVector(Dim_, Index_, off, Vect1_, Contain_, Contain2_);
                      if (CountOfVector(Dim_, Index_, off, Vect1_, Contain_, Contain2_) > Contain_[0])
                        ++Contain_[0];
                    }
                  }
                }
                else
                {
                  if (Vect4_[pos])
                    Vect3_ = Vect4_[pos];
                  else
                    return 0;
                }
              }
              else if (x != 1)
              {
                if (Vect4_[0])
                  Vect3_ = Vect4_[0];
                else
                  return 0;
              }
              break;

            case 6:
              if (pos)
              {
                if (xOff_ == Index_)
                {
                  if (CheckDealloc_ && Vect3_ && Vect3_[0])
                  {
                    Contain2_ = Vect3_[0];
                    Contain_ = Contain2_[0];

                    if (Shared_)
                      *Shared_ = StatusOfVector(Dim_, Index_, off, pos, Vect1_, Contain_);
                      
                    if (Size_)
                      *Size_ = CountOfVector(Dim_, Index_, off, Vect1_, Contain_, Contain2_);

                    if (Max_)
                      *Max_ = Contain_[0];
                  
                    return (Shared_ || Size_);
                  }
                  else
                  {
                    if (Replace_ && !Replacement_)
                    {
                      sz = SizeOfVector(xDim_, xIndex_, LenVect_);
                      Vect2_ = (int**)RawAllocateWith(MEMMATRIX, sizeof(int*) * sz);
                      Replacement_ = InitMatrix(Vect2_, Dim_, off, Index_, LenVect_);
                      Replacement_ = SetRefArrays(Vect2_, Dim_, off, Index_, LenVect_, PosVect_, Error_);
                    }
                  
                    if (Vect3_[pos])
                    {
                      Vect2_ = Vect3_[pos];
                      RemoveReferences(Vect3_, Dim_, off, Index_, LenVect_,
                                       PosVect_, Replacement_, DeAlloc_,
                                       Shared_, Size_, Max_, Error_);
                    }
                    else if (Replacement_)
                    {
                      Vect3_[pos] = Vect2_ = (int**)Replacement_;
                      Contain2_ = Vect3_[0];
                      Contain_ = Contain2_[0];

                      IncCountOfVector(Dim_, Index_, off, Vect1_, Contain_, Contain2_);
                      if (CountOfVector(Dim_, Index_, off, Vect1_, Contain_, Contain2_) > Contain_[0])
                        ++Contain_[0];
                    }
                  }                  
                }
                else
                {
                  if (Vect3_[pos])
                    Vect2_ = Vect3_[pos];
                  else
                    return 0;
                }
              }
              else if (x != 1)              
              {
                if (Vect3_[0])
                  Vect2_ = Vect3_[0];                  
                else
                  return 0;
              }
              break;

            case 7:
              if (pos)
              {
                if (xOff_ == Index_)
                {
                  if (CheckDealloc_ && Vect2_ && Vect2_[0])
                  {
                    Contain2_ = Vect2_;
                    Contain_ = Vect2_[0];

                    if (Shared_)
                      *Shared_ = StatusOfVector(Dim_, Index_, off, pos, Vect1_, Contain_);
                      
                    if (Size_)
                      *Size_ = CountOfVector(Dim_, Index_, off, Vect1_, Contain_, Contain2_);

                    if (Max_)
                      *Max_ = Contain_[0];
                  
                    return (Shared_ || Size_);
                  }
                  else
                  {
                    if (Replace_ && !Replacement_)
                    {
                      sz = SizeOfVector(xDim_, xIndex_, LenVect_);
                      Vect1_ = (int*)RawAllocateWith(MEMMATRIX, sizeof(int) * sz);
                      Replacement_ = InitMatrix(Vect1_, Dim_, off, Index_, LenVect_);
                    }
                  
                    if (Vect2_[pos])
                    {
                      Vect1_ = Vect2_[pos];
                      RemoveReferences(Vect2_, Dim_, off, Index_, LenVect_,
                                       PosVect_, Replacement_, DeAlloc_,
                                       Shared_, Size_, Max_, Error_);
                    }
                    else if (Replacement_)
                    {
                      Vect2_[pos] = Vect1_ = (int*)Replacement_;
                      Contain2_ = Vect2_;
                      Contain_ = Vect2_[0];

                      IncCountOfVector(Dim_, Index_, off, Vect1_, Contain_, Contain2_);
                      if (CountOfVector(Dim_, Index_, off, Vect1_, Contain_, Contain2_) > Contain_[0])
                        ++Contain_[0];
                    }
                  }
                }
                else
                {
                  if (Vect2_[pos])
                    Vect1_ = Vect2_[pos];
                  else
                    return 0;
                }
              }
              else if (x != 1)
              {
                if (Vect2_[0])
                  Vect1_ = Vect2_[0];
                else
                  return 0;
              }
          }
        }
      }
    }
  }
  else if (Dim_ == 1)
  {
    // In calling InitArrayData function for 1D vector,
    // Index argument is assumed to hold the vector length.
    // LenVect argument is assumed to hold deallocated flag and
    // PosVect argument is assumed to hold shallowcopy flag.

    if (CheckDealloc_ && Vect1_)
    {
      if (Shared_)
      {
        *Shared_ = 0;
        
        if (LenVect_ && *LenVect_ == 1)
          *Shared_ = 1;
          
        if (PosVect_ && *PosVect_ == 1)
          *Shared_ = 2;
      }

      if (Max_)
        *Size_ = *Max_ = Vect1_[0];

      return (Shared_ || Size_);
    }
    else
    {
      if (Replacement_ || Replace_)
      {
        ::DeleteArray(Vect1_);

        if (Replace_ && !Replacement_)
        {
          sz = Index_;
          Vect1_ = (int*)RawAllocateWith(MEMMATRIX, sizeof(int) * sz);
          memset(Vect1_, 0, sizeof(int) * sz);
        }
      
        Vect1_ = (int*)Replacement_;
        Vect1_[0] = Index_;
      
        if (PosVect_)
          *PosVect_ = 1;
      }
      else
      {
        if (DeAlloc_)
        {
          ::DeleteArray(Vect1_);
          Vect1_ = NULL;
        
          if (LenVect_)
            *LenVect_ = 1;
        }
        else
          Vect1_[0] = 0;
      }

      if (Shared_)
      {
        *Shared_ = 0;
        
        if (LenVect_ && *LenVect_ == 1)
          *Shared_ = 1;
          
        if (PosVect_ && *PosVect_ == 1)
          *Shared_ = 2;
      }

      if (Max_)
        *Size_ = *Max_ = Vect1_[0];

      Error_ = false;
    }
  }

  return 1;
}

/****************************************************************************/
// PURPOSE:
//   Method to retrieve the deallocated and shared status flags of the
//   specified vector contained within the data matrix and return TRUE if
//   specified vector is deleted (not set to zero size) from the matrix.
//
// PRE:
//   void* Array_  : the generalized multidimensional data matrix
//   int Dim_      : The dimensions of the matrix, value range = [1...8]
//   int Index_    : Index accessing constituent array of the data matrix,
//                   value range = [0...7]
//   int* LenVect_ : Length vector containing length of constituent arrays
//   int* PosVect_ : Position vector containing coordinates into matrix
//   bool& Error_  : Error in array indexing (out of range)
//   int* Shared_  : Set to the deallocation status flag of the container
//                   vector of the specified vector to be deleted.
//                   Valid values include: 0=filled, 1=deallocated, 2=shared
//   int* Size_    : Set to the size in elements of the container vector.
//   int* Max_     : Set to the maximum length of the container vector.
//
int IsDeallocated(void* Array_, int Dim_, int Index_, int* LenVect_,
                  int* PosVect_, bool& Error_, int* Shared_, int* Size_,
                  int* Max_)
{
  int Shval_;  
  int Szval_;
  int Mxval_;

  if (!Shared_)
    Shared_ = &Shval_;

  if (!Size_)
    Size_ = &Szval_;

  if (!Max_)
    Max_ = &Mxval_;

  int ret =
  DeleteVectorHelper(Array_, Dim_, Index_, LenVect_, PosVect_,
                     NULL, false, false, Error_, true, Shared_, Size_, Max_);

  return (ret && *Shared_ == 1);
}

/****************************************************************************/
// PURPOSE:
//   Method to delete the specified element index position in the data matrix
//   and set the element index position to NULL.
//
// PRE:
//   void* Array_  : the generalized multidimensional data matrix
//   int Dim_      : The dimensions of the matrix, value range = [1...8]
//   int Index_    : Index accessing constituent array of the data matrix,
//                   value range = [0...7]
//   int* LenVect_ : Length vector containing length of constituent arrays
//   int* PosVect_ : Position vector containing coordinates into matrix
//   bool& Error_  : Error in array indexing (out of range)
//   int* Shared_  : Set to the deallocation status flag of the container
//                   vector of the specified vector to be deleted.
//                   Valid values include: 0=filled, 1=deallocated, 2=shared
//   int* Size_    : Set to the size in elements of the container vector.
//   int* Max_     : Set to the maximum length of the container vector.
//
int DeleteVector(void* Array_, int Dim_, int Index_, int* LenVect_,
                 int* PosVect_, bool& Error_, int* Shared_, int* Size_,
                 int* Max_)
{
  int ret =
  DeleteVectorHelper(Array_, Dim_, Index_, LenVect_, PosVect_,
                     NULL, false, true, Error_, false, Shared_, Size_, Max_);
  return ret;
}

/****************************************************************************/
// PURPOSE:
//   Method to replace a vector within the data matrix with another vector
//   If the element index within the data matrix doesn't contain a vector
//   then an option to create a new vector and replace it with the newly
//   created vector is available.
//
// PRE:
//   void* Array_  : the generalized multidimensional data matrix
//   int Dim_      : The dimensions of the matrix, value range = [1...8]
//   int Index_    : Index accessing constituent array of the data matrix,
//                   value range = [0...7]
//   int* LenVect_ : Length vector containing length of constituent arrays
//   int* PosVect_ : Position vector containing coordinates into matrix
//
//   void* Replacement_ : The replacement vector to replace the existing
//                        vector in the data matrix with or if the element
//                        position is empty then to set as the new vector.
//   bool Replace_      : If set to true then if Replacement_ is NULL a new
//                        vector is created and then that is used as the
//                        replacement vector, otherwise if false then if
//                        Replacement_ is NULL the element position in the
//                        data matrix is deleted and set to NULL.
//   bool DeAlloc_      : Should actually perform memory deallocation in
//                        data matrix element position instead of setting
//                        vector size to zero.
//   bool& Error_  : Error in array indexing (out of range)
//   int* Shared_  : Set to the deallocation status flag of the container
//                   vector of the specified vector to be deleted.
//                   Valid values include: 0=filled, 1=deallocated, 2=shared
//   int* Size_    : Set to the size in elements of the container vector.
//   int* Max_     : Set to the maximum length of the container vector.
//
int ReplaceVector(void* Array_, int Dim_, int Index_, int* LenVect_,
                  int* PosVect_, void* Replacement_, bool Replace_,
                  bool DeAlloc_, bool& Error_, int* Shared_, int* Size_,
                  int* Max_)
{
  int ret =
  DeleteVectorHelper(Array_, Dim_, Index_, LenVect_, PosVect_,
                     Replacement_, Replace_, DeAlloc_, Error_,
                     false, Shared_, Size_, Max_);
  return ret;
}

/****************************************************************************/
// PURPOSE:
//   Method to return the real vector size within the data matrix given the
//   index into the matrix.
//
// PRE:
//   void* Array_  : the generalized multidimensional data matrix
//   int Dim_      : The dimensions of the matrix, value range = [1...8]
//   int Index_    : Index accessing constituent array of the data matrix,
//                   value range = [0...7]
//   int* LenVect_ : Length vector containing length of constituent arrays
//   int* PosVect_ : Position vector containing coordinates into matrix
//   bool& Error_  : Error in array indexing (out of range)
//
// POST:
//   returns the actual size of the indexed vector in number of elements.
//   The elements of the vector could also be pointer arrays.
//
int RealVectorCount(void* Array_, int Dim_, int Index_,
                    int* LenVect_, int* PosVect_, bool& Error_)
{
  int off = 8 - Dim_; // Maximum dimensions is: 8
  int pos;  
  int sz;
  int x;

  int xIndex_;
  int xOff_;

  xIndex_ = Index_;
  Index_ += off;

  int******** Vect8_ = (Index_ == 0) ? (int********)Array_:NULL;
  int******* Vect7_ = (Index_ == 1) ? (int*******)Array_:NULL;
  int****** Vect6_ = (Index_ == 2) ? (int******)Array_:NULL;
  int***** Vect5_ = (Index_ == 3) ? (int*****)Array_:NULL;
  int**** Vect4_ = (Index_ == 4) ? (int****)Array_:NULL;
  int*** Vect3_ = (Index_ == 5) ? (int***)Array_:NULL;
  int** Vect2_ = (Index_ == 6) ? (int**)Array_:NULL;
  int* Vect1_ = (Index_ == 7) ? (int*)Array_:NULL;
  int** Contain2_;
  int* Contain_;

  if (xIndex_ == 0)
  {
    Vect7_ = (off == 0) ? Vect8_[0]:Vect7_;
    Vect6_ = (off <= 1) ? Vect7_[0]:Vect6_;
    Vect5_ = (off <= 2) ? Vect6_[0]:Vect5_;
    Vect4_ = (off <= 3) ? Vect5_[0]:Vect4_;
    Vect3_ = (off <= 4) ? Vect4_[0]:Vect3_;
    Vect2_ = (off <= 5) ? Vect3_[0]:Vect2_;
    Vect1_ = (off <= 6) ? Vect2_[0]:Vect1_;
    sz = (Dim_ >= 3) ? Vect2_[1][0]:
         (Dim_ == 2) ? Vect2_[0][0]:
                       Vect1_[0];
      
    Error_ = false;
    return sz;
  }
  else
  {
    for (x = 1+off; x <= Index_; x++)
    {
      Error_ = true;
      xOff_ = x - off;
    
      if (xOff_ <= xIndex_ &&
          InRange(xOff_, PosVect_[xOff_], SizeOfVector(Dim_, xOff_-1, LenVect_) - 1))
        pos = PosVect_[xOff_];
      else
        pos = 0;

      if (!pos)
        return 0;
      else
        Error_ = false;
  
      switch (x)
      {
        case 1:
          if (x == Index_)
          {
            if (Vect8_[pos])
            {
              Vect7_ = Vect8_[pos];
              Contain2_ = Vect7_[0][0][0][0][0];
              Vect1_ = Contain_ = Contain2_[0];
              sz = CountOfVector(Dim_, xIndex_, off, Vect1_, Contain_, Contain2_);
              Error_ = false;
              return sz;
            }
          }
          else
          {
            if (Vect8_[pos])
              Vect7_ = Vect8_[pos];
          }
          break;
          
        case 2:
          if (x == Index_)
          {
            if (Vect7_[pos])
            {
              Vect6_ = Vect7_[pos];
              Contain2_ = Vect6_[0][0][0][0];
              Vect1_ = Contain_ = Contain2_[0];
              sz = CountOfVector(Dim_, xIndex_, off, Vect1_, Contain_, Contain2_);
              Error_ = false;
              return sz;
            }
          }
          else
          {
            if (Vect7_[pos])
              Vect6_ = Vect7_[pos];
          }
          break;
      
        case 3:
          if (x == Index_)
          {
            if (Vect6_[pos])
            {
              Vect5_ = Vect6_[pos];
              Contain2_ = Vect5_[0][0][0];
              Vect1_ = Contain_ = Contain2_[0];
              sz = CountOfVector(Dim_, xIndex_, off, Vect1_, Contain_, Contain2_);
              Error_ = false;
              return sz;
            }
          }
          else
          {
            if (Vect6_[pos])
              Vect5_ = Vect6_[pos];
          }
          break;
        
        case 4:
          if (x == Index_)
          {
            if (Vect5_[pos])
            {
              Vect4_ = Vect5_[pos];
              Contain2_ = Vect4_[0][0];
              Vect1_ = Contain_ = Contain2_[0];
              sz = CountOfVector(Dim_, xIndex_, off, Vect1_, Contain_, Contain2_);
              Error_ = false;
              return sz;
            }
          }
          else
          {
            if (Vect5_[pos])
              Vect4_ = Vect5_[pos];
          }
          break;

        case 5:
          if (x == Index_)
          {
            if (Vect4_[pos])
            {
              Vect3_ = Vect4_[pos];
              Contain2_ = Vect3_[0];
              Vect1_ = Contain_ = Contain2_[0];
              sz = CountOfVector(Dim_, xIndex_, off, Vect1_, Contain_, Contain2_);
              Error_ = false;
              return sz;
            }
          }
          else
          {
            if (Vect4_[pos])
              Vect3_ = Vect4_[pos];
          }
          break;

        case 6:
          if (x == Index_)
          {
            if (Vect3_[pos])
            {
              Vect2_ = Vect3_[pos];
              Contain2_ = Vect2_;
              Vect1_ = Contain_ = Vect2_[0];
              sz = CountOfVector(Dim_, xIndex_, off, Vect1_, Contain_, Contain2_);
              Error_ = false;
              return sz;
            }
          }
          else
          {
            if (Vect3_[pos])
              Vect2_ = Vect3_[pos];
          }
          break;

        case 7:
          if (x == Index_)
          {
            if (Vect2_[pos])
            {
              Vect1_ = Vect2_[pos];
              sz = Vect1_[0];
              Error_ = false;
              return sz;
            }
          }
          else
          {
            if (Vect2_[pos])
            {
              Vect1_ = Vect2_[pos];
              sz = Vect1_[0];
              Error_ = false;
              return sz;
            }
          }
          break;
      }
    }
  }

  return 0;
}

/****************************************************************************/
// PURPOSE:
//   Method to return the real vector maximum length within the data matrix
//   given the index into the matrix.
//
// PRE:
//   void* Array_  : the generalized multidimensional data matrix
//   int Dim_      : The dimensions of the matrix, value range = [1...8]
//   int Index_    : Index accessing constituent array of the data matrix,
//                   value range = [0...7]
//   int* LenVect_ : Length vector containing length of constituent arrays
//   int* PosVect_ : Position vector containing coordinates into matrix
//   bool& Error_  : Error in array indexing (out of range)
//
// POST:
//   returns the actual size of the indexed vector in number of elements.
//   The elements of the vector could also be pointer arrays.
//
int RealVectorMax(void* Array_, int Dim_, int Index_,
                  int* LenVect_, int* PosVect_, bool& Error_)
{
  int off = 8 - Dim_; // Maximum dimensions is: 8
  int pos;  
  int sz;
  int x;

  int xIndex_;
  int xOff_;

  xIndex_ = Index_;
  Index_ += off;

  int******** Vect8_ = (Index_ == 0) ? (int********)Array_:NULL;
  int******* Vect7_ = (Index_ == 1) ? (int*******)Array_:NULL;
  int****** Vect6_ = (Index_ == 2) ? (int******)Array_:NULL;
  int***** Vect5_ = (Index_ == 3) ? (int*****)Array_:NULL;
  int**** Vect4_ = (Index_ == 4) ? (int****)Array_:NULL;
  int*** Vect3_ = (Index_ == 5) ? (int***)Array_:NULL;
  int** Vect2_ = (Index_ == 6) ? (int**)Array_:NULL;
  int* Vect1_ = (Index_ == 7) ? (int*)Array_:NULL;

  if (xIndex_ == 0)
  {
    Vect7_ = (off == 0) ? Vect8_[0]:Vect7_;
    Vect6_ = (off <= 1) ? Vect7_[0]:Vect6_;
    Vect5_ = (off <= 2) ? Vect6_[0]:Vect5_;
    Vect4_ = (off <= 3) ? Vect5_[0]:Vect4_;
    Vect3_ = (off <= 4) ? Vect4_[0]:Vect3_;
    Vect2_ = (off <= 5) ? Vect3_[0]:Vect2_;
    Vect1_ = (off <= 6) ? Vect2_[0]:Vect1_;
    
    sz = Vect1_[0];
    Error_ = false;
    return sz;
  }
  else
  {
    for (x = 1+off; x <= Index_; x++)
    {
      Error_ = true;
      xOff_ = x - off;
    
      if (xOff_ <= xIndex_ &&
          InRange(xOff_, PosVect_[xOff_], SizeOfVector(Dim_, xOff_-1, LenVect_) - 1))
        pos = PosVect_[xOff_];
      else
        pos = 0;

      if (!pos)
        return 0;
      else
        Error_ = false;
  
      switch (x)
      {
        case 1:
          if (x == Index_)
          {
            if (Vect8_[pos])
            {
              Vect7_ = Vect8_[pos];
              sz = Vect7_[0][0][0][0][0][0][0];
              Error_ = false;
              return sz;
            }
          }
          else
          {
            if (Vect8_[pos])
              Vect7_ = Vect8_[pos];
          }
          break;
          
        case 2:
          if (x == Index_)
          {
            if (Vect7_[pos])
            {
              Vect6_ = Vect7_[pos];
              sz = Vect6_[0][0][0][0][0][0];
              Error_ = false;
              return sz;
            }
          }
          else
          {
            if (Vect7_[pos])
              Vect6_ = Vect7_[pos];
          }
          break;
      
        case 3:
          if (x == Index_)
          {
            if (Vect6_[pos])
            {
              Vect5_ = Vect6_[pos];
              sz = Vect5_[0][0][0][0][0];
              Error_ = false;
              return sz;
            }
          }
          else
          {
            if (Vect6_[pos])
              Vect5_ = Vect6_[pos];
          }
          break;
        
        case 4:
          if (x == Index_)
          {
            if (Vect5_[pos])
            {
              Vect4_ = Vect5_[pos];
              sz = Vect4_[0][0][0][0];
              Error_ = false;
              return sz;
            }
          }
          else
          {
            if (Vect5_[pos])
              Vect4_ = Vect5_[pos];
          }
          break;

        case 5:
          if (x == Index_)
          {
            if (Vect4_[pos])
            {
              Vect3_ = Vect4_[pos];
              sz = Vect3_[0][0][0];
              Error_ = false;
              return sz;
            }
          }
          else
          {
            if (Vect4_[pos])
              Vect3_ = Vect4_[pos];
          }
          break;

        case 6:
          if (x == Index_)
          {
            if (Vect3_[pos])
            {
              Vect2_ = Vect3_[pos];
              sz = Vect2_[0][0];
              Error_ = false;
              return sz;
            }
          }
          else
          {
            if (Vect3_[pos])
              Vect2_ = Vect3_[pos];
          }
          break;

        case 7:
          if (x == Index_)
          {
            if (Vect2_[pos])
            {
              Vect1_ = Vect2_[pos];
              sz = Vect1_[0];
              Error_ = false;
              return sz;
            }
          }
          else
          {
            if (Vect2_[pos])
            {
              Vect1_ = Vect2_[pos];
              sz = Vect1_[0];
              Error_ = false;
              return sz;
            }
          }
          break;
      }
    }
  }

  return 0;
}

/****************************************************************************/
// PURPOSE:
//   Method to retrieve the parent reference array given the index into the
//   general data matrix, valid only for matrix dim > 2.
//
// PRE:
//   void* Array_     : the generalized multidimensional data matrix
//   int Dim_         : The dimensions of the matrix, value range = [1...8]
//   int Index_       : Index accessing constituent array of the data matrix,
//                      value range = [0...7]
//   int* LenVect_    : Length vector containing length of constituent arrays
//   int* PosVect_    : Position vector containing coordinates into matrix
//   int* ParentRefs_ : Parent reference array of specified constituent vector
//   bool& Error_     : Error in array indexing (out of range)
//
// POST:
//   If Dim_ > 3 then returns the parent reference array as specified by the
//   Index into the general data matrix, otherwise returns NULL.
//
void* ParentReference(void* Array_, int Dim_, int Index_,
                      int* LenVect_, int* PosVect_, int* ParentRefs_,
                      bool& Error_)
{
  if (Dim_ < 3)
    return NULL;

  int off = 8 - Dim_; // Maximum dimensions is: 8
  int pos;  
  int sz;
  int x;

  int xIndex_;
  int xOff_;
  int xDim_;  

  xIndex_ = Index_;
  Index_ += off;
  xDim_ = Dim_;
  Dim_ += off;
  
  int******** Vect8_ = (Index_ == 0) ? (int********)Array_:NULL;;
  int******* Vect7_ = (Index_ == 1) ? (int*******)Array_:NULL;;
  int****** Vect6_ = (Index_ == 2) ? (int******)Array_:NULL;
  int***** Vect5_ = (Index_ == 3) ? (int*****)Array_:NULL;
  int**** Vect4_ = (Index_ == 4) ? (int****)Array_:NULL;
  int*** Vect3_ = (Index_ == 5) ? (int***)Array_:NULL;
  int** Vect2_ = (Index_ == 6) ? (int**)Array_:NULL;
  int* Vect1_ = (Index_ == 7) ? (int*)Array_:NULL;  

  if (xIndex_ != 0)
  {
    for (x = 1+off; x <= Index_; x++)
    {
      Error_ = true;
      xOff_ = x - off;  // real index position into matrix minus offsets
                        // x is index position with offsets included
    
      if (xOff_ <= xIndex_ &&
          InRange(xOff_, PosVect_[xOff_], SizeOfVector(xDim_, xOff_-1, LenVect_) - 1))
        pos = PosVect_[xOff_];
      else
        pos = 0;

      if (!pos)
        return 0;
      else
        Error_ = false;

      switch (x)
      {
        case 1:
          if (x == Index_)
          {
            if (Vect8_[pos])
            {
              Vect7_ = Vect8_[pos];
              ParentRefs_ =
              GetParentReference(ParentRefs_, Dim_, off, Index_, LenVect_, PosVect_, Vect7_, Error_);
              return ParentRefs_;
            }
          }
          else
          {
            if (Vect8_[pos])
              Vect7_ = Vect8_[pos];
          }
          break;
          
        case 2:
          if (x == Index_)
          {
            if (Vect7_[pos])
            {
              Vect6_ = Vect7_[pos];
              ParentRefs_ =
              GetParentReference(ParentRefs_, Dim_, off, Index_, LenVect_, PosVect_, Vect6_, Error_);
              return ParentRefs_;
            }
          }
          else
          {
            if (Vect7_[pos])
              Vect6_ = Vect7_[pos];
          }
          break;
      
        case 3:
          if (x == Index_)
          {
            if (Vect6_[pos])
            {
              Vect5_ = Vect6_[pos];
              ParentRefs_ =
              GetParentReference(ParentRefs_, Dim_, off, Index_, LenVect_, PosVect_, Vect5_, Error_);
              return ParentRefs_;              
            }
          }
          else
          {
            if (Vect6_[pos])
              Vect5_ = Vect6_[pos];
          }
          break;
        
        case 4:
          if (x == Index_)
          {
            if (Vect5_[pos])
            {
              Vect4_ = Vect5_[pos];
              ParentRefs_ =
              GetParentReference(ParentRefs_, Dim_, off, Index_, LenVect_, PosVect_, Vect4_, Error_);
              return ParentRefs_;
            }
          }
          else
          {
            if (Vect5_[pos])
              Vect4_ = Vect5_[pos];
          }
          break;

        case 5:
          if (x == Index_)
          {
            if (Vect4_[pos])
            {
              Vect3_ = Vect4_[pos];
              ParentRefs_ =
              GetParentReference(ParentRefs_, Dim_, off, Index_, LenVect_, PosVect_, Vect3_, Error_);
              return ParentRefs_;
            }
          }
          else
          {
            if (Vect4_[pos])
              Vect3_ = Vect4_[pos];
          }
          break;

        case 6:
          if (x == Index_)
          {
            if (Vect3_[pos])
            {
              Vect2_ = Vect3_[pos];
              ParentRefs_ =
              GetParentReference(ParentRefs_, Dim_, off, Index_, LenVect_, PosVect_, Vect2_, Error_);
              return ParentRefs_;              
            }
          }
          else
          {
            if (Vect3_[pos])
              Vect2_ = Vect3_[pos];
          }
          break;

        case 7:
          if (x == Index_)
          {
            if (Vect2_[pos])
            {
              Vect1_ = Vect2_[pos];
              ParentRefs_ =
              GetParentReference(ParentRefs_, Dim_, off, Index_, LenVect_, PosVect_, Vect1_, Error_);
              return ParentRefs_;              
            }
          }
          else
          {
            if (Vect2_[pos])
            {
              Vect1_ = Vect2_[pos];
              ParentRefs_ =
              GetParentReference(ParentRefs_, Dim_, off, Index_, LenVect_, PosVect_, Vect1_, Error_);
              return ParentRefs_;
            }
          }
          break;        
      }
    }
  }

  return NULL;
}

/****************************************************************************/
// PURPOSE:
//   Method to read integer data from the general (void*) multidimensional
//   data matrix.
//
// PRE:
//   void* Array  : Data matrix to initialize
//   int Dim      : Dimensions of matrix
//   int Index    : Index position into matrix of arrays
//   int* LenVect : Length vector containing length of constituent arrays
//   int* PosVect : Position vector containing coordinates into matrix
//   bool& Error_ : Error in array indexing (out of range)
//
// POST:
//   Returns integer data in the specified positions from the general data
//   matrix.
//
int ReadFromArray(void* Array_, int Dim_, int Index_,
                  int* LenVect_, int* PosVect_, bool& Error_)
{
  int******** Vect8_;
  int******* Vect7_;
  int****** Vect6_;
  int***** Vect5_;
  int**** Vect4_;
  int*** Vect3_;
  int** Vect2_;
  int* Vect1_;

  int off = 8 - Dim_; // Maximum dimensions is: 8
  int pos;  
  int x;

  int xIndex_;
  int xDim_;
  int xOff_;
  int ret;
  Index_ = Dim_ - 1;

  if (2 <= Dim_ && Dim_ <= 8)
  {
    if (Index_ != 0)
    {
      if (Dim_ == 2)
      {
        x = 1;
        if (InRange(x, PosVect_[x], SizeOfVector(Dim_, x-1, LenVect_) - 1))
          pos = PosVect_[x];
        else
          pos = 0;

        if (pos && Array_)
        {      
          Vect2_ = (int**)Array_;
          Vect1_ = Vect2_[pos];
          ++x;

          if (InRange(x, PosVect_[x], SizeOfVector(Dim_, x-1, LenVect_) - 1))
            pos = PosVect_[x];
          else
            pos = 0;          

          // Setting vector size
          if (pos && Vect1_)
          {
            ret = Vect1_[pos];
            Error_ = false;
            return ret;
          }
          else
          {
            Error_ = true;
            return 0;
          }
        }
        else
        {
          Error_ = true;
          return 0;
        }
      }
      else
      {
        Vect8_ = (off == 0) ? (int********)Array_:NULL;
        Vect7_ = (off == 1) ? (int*******)Array_:NULL;
        Vect6_ = (off == 2) ? (int******)Array_:NULL;
        Vect5_ = (off == 3) ? (int*****)Array_:NULL;
        Vect4_ = (off == 4) ? (int****)Array_:NULL;
        Vect3_ = (off == 5) ? (int***)Array_:NULL;
        Vect2_ = (off == 6) ? (int**)Array_:NULL;      

        xDim_ = Dim_;
        Dim_ += off;
        xIndex_ = Index_;
        Index_ += off;

        for (x = 1; x < xDim_; x++)
        {
          xOff_ = x + off;

          if ((Error_ = (x <= xIndex_)) &&
              InRange(x, PosVect_[x], SizeOfVector(xDim_, x-1, LenVect_) - 1))
            pos = PosVect_[x];
          else
            pos = 0;

          if (!pos && Error_)
          {
            Error_ = true;
            return 0;
          }
          else
            Error_ = false;
                        
          switch (xOff_)
          {
            case 1:
              if (pos)
              {
                if (xOff_ == Index_)
                {
                  if (Vect8_[pos])
                  {
                    Vect7_ = Vect8_[pos];

                    if (InRange(x+1, PosVect_[x+1], SizeOfVector(xDim_, x, LenVect_) - 1))
                      pos = PosVect_[x+1];
                    else
                      pos = 0;

                    if (pos && Vect7_[pos])
                    {
                      Error_ = false;
                      return 0;
                    }
                    else
                      Error_ = true;                    
                  }
                  else
                  {
                    Error_ = true;
                    return 0;
                  }
                }
                else
                {
                  if (Vect8_[pos])
                    Vect7_ = Vect8_[pos];
                  else
                  {
                    Error_ = true;
                    return 0;
                  }
                }
              }
              else
              {
                Error_ = true;
                return 0;              
              }
              break;

            case 2:
              if (pos)
              {
                if (xOff_ == Index_)
                {
                  if (Vect7_[pos])
                  {
                    Vect6_ = Vect7_[pos];

                    if (InRange(x+1, PosVect_[x+1], SizeOfVector(xDim_, x, LenVect_) - 1))
                      pos = PosVect_[x+1];
                    else
                      pos = 0;

                    if (pos && Vect6_[pos])
                    {
                      Error_ = false;
                      return 0;
                    }
                    else
                      Error_ = true;                    
                  }
                  else
                  {
                    Error_ = true;
                    return 0;
                  }
                }
                else
                {
                  if (Vect7_[pos])
                    Vect6_ = Vect7_[pos];
                  else
                  {
                    Error_ = true;
                    return 0;
                  }
                }
              }
              else
              {
                Error_ = true;
                return 0;
              }
              break;
              
            case 3:
              if (pos)
              {
                if (xOff_ == Index_)
                {
                  if (Vect6_[pos])
                  {
                    Vect5_ = Vect6_[pos];

                    if (InRange(x+1, PosVect_[x+1], SizeOfVector(xDim_, x, LenVect_) - 1))
                      pos = PosVect_[x+1];
                    else
                      pos = 0;

                    if (pos && Vect5_[pos])
                    {
                      Error_ = false;
                      return 0;
                    }
                    else
                      Error_ = true;                    
                  }
                  else
                  {
                    Error_ = true;
                    return 0;
                  }
                }
                else
                {
                  if (Vect6_[pos])
                    Vect5_ = Vect6_[pos];
                  else
                  {
                    Error_ = true;
                    return 0;
                  }
                }
              }
              else
              {
                Error_ = true;
                return 0;
              }            
              break;
            
            case 4:
              if (pos)
              {
                if (xOff_ == Index_)
                {
                  if (Vect5_[pos])
                  {
                    Vect4_ = Vect5_[pos];

                    if (InRange(x+1, PosVect_[x+1], SizeOfVector(xDim_, x, LenVect_) - 1))
                      pos = PosVect_[x+1];
                    else
                      pos = 0;

                    if (pos && Vect4_[pos])
                    {
                      Error_ = false;
                      return 0;
                    }
                    else
                      Error_ = true;                    
                  }
                  else
                  {
                    Error_ = true;
                    return 0;
                  }
                }
                else
                {
                  if (Vect5_[pos])
                    Vect4_ = Vect5_[pos];
                  else
                  {
                    Error_ = true;
                    return 0;
                  }
                }
              }
              else
              {
                Error_ = true;
                return 0;
              }            
              break;
            
            case 5:
              if (pos)
              {
                if (xOff_ == Index_)
                {
                  if (Vect4_[pos])
                  {
                    Vect3_ = Vect4_[pos];

                    if (InRange(x+1, PosVect_[x+1], SizeOfVector(xDim_, x, LenVect_) - 1))
                      pos = PosVect_[x+1];
                    else
                      pos = 0;

                    if (pos && Vect3_[pos])
                    {
                      Error_ = false;
                      return 0;
                    }
                    else
                      Error_ = true;                    
                  }
                  else
                  {
                    Error_ = true;
                    return 0;
                  }
                }
                else
                {
                  if (Vect4_[pos])
                    Vect3_ = Vect4_[pos];
                  else
                  {
                    Error_ = true;
                    return 0;
                  }
                }
              }
              else
              {
                Error_ = true;
                return 0;
              }
              break;

            case 6:
              if (pos)
              {
                if (xOff_ == Index_)
                {
                  if (Vect3_[pos])
                  {
                    Vect2_ = Vect3_[pos];

                    if (InRange(x+1, PosVect_[x+1], SizeOfVector(xDim_, x, LenVect_) - 1))
                      pos = PosVect_[x+1];
                    else
                      pos = 0;

                    if (pos && Vect2_[pos])
                    {
                      Error_ = false;
                      return 0;
                    }
                    else
                      Error_ = true;                    
                  }
                  else
                  {
                    Error_ = true;
                    return 0;
                  }
                }
                else
                {
                  if (Vect3_[pos])
                    Vect2_ = Vect3_[pos];
                  else
                  {
                    Error_ = true;
                    return 0;
                  }
                }
              }
              else
              {
                Error_ = true;
                return 0;
              }            
              break;

            case 7:
              if (pos)
              {
                if (xOff_ == Index_)
                {
                  if (Vect2_[pos])
                  {
                    Vect1_ = Vect2_[pos];

                    if (InRange(1, PosVect_[x+1], SizeOfVector(xDim_, x, LenVect_) - 1))
                      pos = PosVect_[x+1];
                    else
                      pos = 0;

                    if (pos)
                    {
                      ret = Vect1_[pos];
                      Error_ = false;
                      return ret;
                    }
                    else
                      Error_ = true;
                  }
                  else
                  {
                    Error_ = true;
                    return 0;
                  }
                }
                else
                {
                  if (Vect2_[pos])
                    Vect1_ = Vect2_[pos];
                  else
                  {
                    Error_ = true;
                    return 0;
                  }
                }
              }
              else
              {
                Error_ = true;
                return 0;
              }
          }
        }
      }
    }
  }
  else if (Dim_ == 1)
  {
    // In calling InitArrayData function for 1D vector,
    // Index argument is assumed to hold the array index position.
    if (InRange(1, PosVect_[1], SizeOfVector(xDim_, 0, LenVect_) - 1))
      pos = PosVect_[1];
    else
      pos = 0;
    
    if (pos)
    {
      ret = Vect1_[pos];
      Error_ = false;
      return ret;
    }
    else
      Error_ = true;
  }

  return 0;
}

/****************************************************************************/
// PURPOSE:
//   Method to write into the general (void*) multidimensional data matrix
//
// PRE:
//   void* Array  : Data matrix to initialize
//   int Dim      : Dimensions of matrix
//   int Index    : Index position into matrix of arrays
//   int* LenVect : Length vector containing length of constituent arrays
//   int* PosVect : Position vector containing coordinates into matrix
//   int Data_    : Integer data to assign into the data matrix.
//   bool& Error_ : Error in array indexing (out of range)
//
// POST:
//   Returns passed data matrix assigned with the specified integer data.
//
void* WriteToArray(void* Array_, int Dim_, int Index_,
                   int* LenVect_, int* PosVect_, int Data_, bool& Error_)
{
  int******** Vect8_;
  int******* Vect7_;
  int****** Vect6_;
  int***** Vect5_;
  int**** Vect4_;
  int*** Vect3_;
  int** Vect2_;
  int* Vect1_;

  int off = 8 - Dim_; // Maximum dimensions is: 8
  int pos;  
  int x;

  int xIndex_;
  int xDim_;
  int xOff_;
  Index_ = Dim_ - 1;

  if (2 <= Dim_ && Dim_ <= 8)
  {
    if (Index_ != 0)
    {
      if (Dim_ == 2)
      {
        x = 1;
        if (InRange(x, PosVect_[x], SizeOfVector(Dim_, x-1, LenVect_) - 1))
          pos = PosVect_[x];
        else
          pos = 0;

        if (pos && Array_)
        {      
          Vect2_ = (int**)Array_;
          Vect1_ = Vect2_[pos];
          ++x;

          if (InRange(x, PosVect_[x], SizeOfVector(Dim_, x-1, LenVect_) - 1))
            pos = PosVect_[x];
          else
            pos = 0;          

          // Setting vector size
          if (pos && Vect1_)
          {
            Vect1_[pos] = Data_;
            Error_ = false;
            return Array_;
          }
          else
          {
            Error_ = true;
            return Array_;
          }
        }
        else
        {
          Error_ = true;
          return Array_;
        }
      }
      else
      {
        Vect8_ = (off == 0) ? (int********)Array_:NULL;
        Vect7_ = (off == 1) ? (int*******)Array_:NULL;
        Vect6_ = (off == 2) ? (int******)Array_:NULL;
        Vect5_ = (off == 3) ? (int*****)Array_:NULL;
        Vect4_ = (off == 4) ? (int****)Array_:NULL;
        Vect3_ = (off == 5) ? (int***)Array_:NULL;
        Vect2_ = (off == 6) ? (int**)Array_:NULL;      

        xDim_ = Dim_;
        Dim_ += off;
        xIndex_ = Index_;
        Index_ += off;

        for (x = 1; x < xDim_; x++)
        {
          xOff_ = x + off;

          if ((Error_ = (x <= xIndex_)) &&
              InRange(x, PosVect_[x], SizeOfVector(xDim_, x-1, LenVect_) - 1))
            pos = PosVect_[x];
          else
            pos = 0;

          if (!pos && Error_)
          {
            Error_ = true;
            return Array_;
          }
          else
            Error_ = false;
                        
          switch (xOff_)
          {
            case 1:
              if (pos)
              {
                if (xOff_ == Index_)
                {
                  if (Vect8_[pos])
                  {
                    Vect7_ = Vect8_[pos];

                    if (InRange(x+1, PosVect_[x+1], SizeOfVector(xDim_, x, LenVect_) - 1))
                      pos = PosVect_[x+1];
                    else
                      pos = 0;

                    if (pos && Vect7_[pos])
                    {
                      Error_ = false;
                      return Array_;
                    }
                    else
                      Error_ = true;                    
                  }
                  else
                  {
                    Error_ = true;
                    return Array_;
                  }
                }
                else
                {
                  if (Vect8_[pos])
                    Vect7_ = Vect8_[pos];
                  else
                  {
                    Error_ = true;
                    return Array_;
                  }
                }
              }
              else
              {
                Error_ = true;
                return Array_;              
              }
              break;

            case 2:
              if (pos)
              {
                if (xOff_ == Index_)
                {
                  if (Vect7_[pos])
                  {
                    Vect6_ = Vect7_[pos];

                    if (InRange(x+1, PosVect_[x+1], SizeOfVector(xDim_, x, LenVect_) - 1))
                      pos = PosVect_[x+1];
                    else
                      pos = 0;

                    if (pos && Vect6_[pos])
                    {
                      Error_ = false;
                      return Array_;
                    }
                    else
                      Error_ = true;                    
                  }
                  else
                  {
                    Error_ = true;
                    return Array_;
                  }
                }
                else
                {
                  if (Vect7_[pos])
                    Vect6_ = Vect7_[pos];
                  else
                  {
                    Error_ = true;
                    return Array_;
                  }
                }
              }
              else
              {
                Error_ = true;
                return Array_;
              }
              break;
              
            case 3:
              if (pos)
              {
                if (xOff_ == Index_)
                {
                  if (Vect6_[pos])
                  {
                    Vect5_ = Vect6_[pos];

                    if (InRange(x+1, PosVect_[x+1], SizeOfVector(xDim_, x, LenVect_) - 1))
                      pos = PosVect_[x+1];
                    else
                      pos = 0;

                    if (pos && Vect5_[pos])
                    {
                      Error_ = false;
                      return Array_;
                    }
                    else
                      Error_ = true;                    
                  }
                  else
                  {
                    Error_ = true;
                    return Array_;
                  }
                }
                else
                {
                  if (Vect6_[pos])
                    Vect5_ = Vect6_[pos];
                  else
                  {
                    Error_ = true;
                    return Array_;
                  }
                }
              }
              else
              {
                Error_ = true;
                return Array_;
              }            
              break;
            
            case 4:
              if (pos)
              {
                if (xOff_ == Index_)
                {
                  if (Vect5_[pos])
                  {
                    Vect4_ = Vect5_[pos];

                    if (InRange(x+1, PosVect_[x+1], SizeOfVector(xDim_, x, LenVect_) - 1))
                      pos = PosVect_[x+1];
                    else
                      pos = 0;

                    if (pos && Vect4_[pos])
                    {
                      Error_ = false;
                      return Array_;
                    }
                    else
                      Error_ = true;                    
                  }
                  else
                  {
                    Error_ = true;
                    return Array_;
                  }
                }
                else
                {
                  if (Vect5_[pos])
                    Vect4_ = Vect5_[pos];
                  else
                  {
                    Error_ = true;
                    return Array_;
                  }
                }
              }
              else
              {
                Error_ = true;
                return Array_;
              }            
              break;
            
            case 5:
              if (pos)
              {
                if (xOff_ == Index_)
                {
                  if (Vect4_[pos])
                  {
                    Vect3_ = Vect4_[pos];

                    if (InRange(x+1, PosVect_[x+1], SizeOfVector(xDim_, x, LenVect_) - 1))
                      pos = PosVect_[x+1];
                    else
                      pos = 0;

                    if (pos && Vect3_[pos])
                    {
                      Error_ = false;
                      return Array_;
                    }
                    else
                      Error_ = true;                    
                  }
                  else
                  {
                    Error_ = true;
                    return Array_;
                  }
                }
                else
                {
                  if (Vect4_[pos])
                    Vect3_ = Vect4_[pos];
                  else
                  {
                    Error_ = true;
                    return Array_;
                  }
                }
              }
              else
              {
                Error_ = true;
                return Array_;
              }
              break;

            case 6:
              if (pos)
              {
                if (xOff_ == Index_)
                {
                  if (Vect3_[pos])
                  {
                    Vect2_ = Vect3_[pos];

                    if (InRange(x+1, PosVect_[x+1], SizeOfVector(xDim_, x, LenVect_) - 1))
                      pos = PosVect_[x+1];
                    else
                      pos = 0;

                    if (pos && Vect2_[pos])
                    {
                      Error_ = false;
                      return Array_;
                    }
                    else
                      Error_ = true;                    
                  }
                  else
                  {
                    Error_ = true;
                    return Array_;
                  }
                }
                else
                {
                  if (Vect3_[pos])
                    Vect2_ = Vect3_[pos];
                  else
                  {
                    Error_ = true;
                    return Array_;
                  }
                }
              }
              else
              {
                Error_ = true;
                return Array_;
              }            
              break;

            case 7:
              if (pos)
              {
                if (xOff_ == Index_)
                {
                  if (Vect2_[pos])
                  {
                    Vect1_ = Vect2_[pos];

                    if (InRange(1, PosVect_[x+1], SizeOfVector(xDim_, x, LenVect_) - 1))
                      pos = PosVect_[x+1];
                    else
                      pos = 0;

                    if (pos)
                    {
                      Vect1_[pos] = Data_;
                      Error_ = false;
                      return Array_;
                    }
                    else
                      Error_ = true;
                  }
                  else
                  {
                    Error_ = true;
                    return Array_;
                  }
                }
                else
                {
                  if (Vect2_[pos])
                    Vect1_ = Vect2_[pos];
                  else
                  {
                    Error_ = true;
                    return Array_;
                  }
                }
              }
              else
              {
                Error_ = true;
                return Array_;
              }
          }
        }
      }
    }
  }
  else if (Dim_ == 1)
  {
    // In calling InitArrayData function for 1D vector,
    // Index argument is assumed to hold the array index position.
    if (InRange(1, PosVect_[1], SizeOfVector(xDim_, 0, LenVect_) - 1))
      pos = PosVect_[1];
    else
      pos = 0;
    
    if (pos)
    {
      Vect1_[pos] = Data_;
      Array_ = Vect1_;      
      Error_ = false;
      return Array_;
    }
    else
      Error_ = true;
  }

  return Array_;
}

/****************************************************************************/
void* DestroyArrayData(void* Array_, int Dim_, int* LenVect_)
{
  int******** Vect8_;
  int******* Vect7_;
  int****** Vect6_;
  int***** Vect5_;
  int**** Vect4_;
  int*** Vect3_;
  int** Vect2_;
  int* Vect1_;

  int off = 8 - Dim_; // Maximum dimensions is: 8
  int pos;  
  int sz;
  int max;
  int x;

  int xIndex_;
  int Index_ = 1;
  int xDim_;
  int xOff_;

  bool RefMatrixDone_;

  if (2 <= Dim_ && Dim_ <= 8)
  {
      if (Dim_ == 2)
      {
        max = SizeOfVector(Dim_, 0, LenVect_);

        if (max)
        {
          Vect2_ = (int**)Array_;
          for (x = 0; x < max; x++)
            if (Vect2_[x])
              ::DeleteArray(Vect2_[x]);

          ::DeleteArray(Vect2_);
          Array_ = NULL;
        }
        else
          return NULL;
      }
      else
      {      
        Vect8_ = (off == 0) ? (int********)Array_:NULL;
        Vect7_ = (off == 1) ? (int*******)Array_:NULL;
        Vect6_ = (off == 2) ? (int******)Array_:NULL;
        Vect5_ = (off == 3) ? (int*****)Array_:NULL;
        Vect4_ = (off == 4) ? (int****)Array_:NULL;
        Vect3_ = (off == 5) ? (int***)Array_:NULL;
        Vect2_ = (off == 6) ? (int**)Array_:NULL;
      
        RefMatrixDone_ = false;
        sz = SizeOfVector(Dim_, Index_, LenVect_);

        xDim_ = Dim_;
        Dim_ += off;
        xIndex_ = Index_;
        Index_ += off;

        // RemoveArray(void* Array_, int Dim_, int Off_, int MatrixIndex_, int* LenVect_);
          switch (off)
          {
            case 0:
              break;

            case 1:
              break;
              
            case 2:
              break;
            
            case 3:
              break;
            
            case 4:
              break;

            case 5:
              break;

            case 6:
              break;
          }
      }
  }
  else if (Dim_ == 1)
  {
    Vect1_ = (int*)Array_;    
    ::DeleteArray(Vect1_);
    Array_ = NULL;
  }

  return Array_;
}

/****************************************************************************/
// PURPOSE:
//   Method to intialize general (void*) multidimensional data matrix
//
// PRE:
//   void* Array  : Data matrix to initialize
//   int Dim      : Dimensions of matrix
//   int Index    : Index position into matrix of arrays
//   int* LenVect : Length vector containing length of constituent arrays
//   int* PosVect : Position vector containing coordinates into matrix
//   bool& Error_ : Error in array indexing (out of range)
//
// POST:
//   Returns initialized data matrix of given dimensions and of given lengths
//
void* InitArrayData(void* Array_, int Dim_, int Index_,
                    int* LenVect_, int* PosVect_, bool& Error_)
{
  int******** Vect8_;
  int******* Vect7_;
  int****** Vect6_;
  int***** Vect5_;
  int**** Vect4_;
  int*** Vect3_;
  int** Vect2_;
  int* Vect1_;
  int** Contain2_;
  int* Contain_;

  int off = 8 - Dim_; // Maximum dimensions is: 8
  int pos;
  int max;
  int sz;
  int x;

  int xIndex_;
  int xDim_;
  int xOff_;

  bool RefMatrixDone_;
  
  // Contain_[pos] = 1;    // deleted
  // Contain_[pos] = 2;    // shared
  // SizeOfVector(Dim_, x, LenVect_)
  // CountOfVector(Dim_, x, Off_, Vect1_, Contain_, Contain2_)
  // StatusOfVector(Dim_, x, Off_, y, Vect1_, Contain_)
  // SetStatusOfVector(Dim_, x, Off_, y, stat, Contain_)

  if (2 <= Dim_ && Dim_ <= 8 &&
      0 <= Index_ && Index_ < Dim_)
  {
    sz = SizeOfVector(Dim_, Index_, LenVect_);
    if (sz < 2)
      return NULL;

    if (Index_ == 0)
    {
      Vect8_ = (off == 0) ?
      (int********)RawAllocateWith(MEMMATRIX, sizeof(int*******) * LenVect_[0]+1):
      NULL;
    
      Vect7_ = (off == 1) ?
      (int*******)RawAllocateWith(MEMMATRIX, sizeof(int******) * LenVect_[0]+1):
      (off < 1) ? (Vect8_[0] = (int*******)RawAllocateWith(MEMMATRIX, sizeof(int******))):
      NULL;
    
      Vect6_ = (off == 2) ?
      (int******)RawAllocateWith(MEMMATRIX, sizeof(int*****) * LenVect_[0]+1):
      (off < 2) ? (Vect7_[0] = (int******)RawAllocateWith(MEMMATRIX, sizeof(int*****))):
      NULL;
      
      Vect5_ = (off == 3) ?
      (int*****)RawAllocateWith(MEMMATRIX, sizeof(int****) * LenVect_[0]+1):
      (off < 3) ? (Vect6_[0] = (int*****)RawAllocateWith(MEMMATRIX, sizeof(int****))):
      NULL;
      
      Vect4_ = (off == 4) ?
      (int****)RawAllocateWith(MEMMATRIX, sizeof(int***) * LenVect_[0]+1):
      (off < 4) ? (Vect5_[0] = (int****)RawAllocateWith(MEMMATRIX, sizeof(int***))):
      NULL;
      
      Vect3_ = (off == 5) ?
      (int***)RawAllocateWith(MEMMATRIX, sizeof(int**) * LenVect_[0]+1):
      (off < 5) ? (Vect4_[0] = (int***)RawAllocateWith(MEMMATRIX, sizeof(int**))):
      NULL;

      max = (off == 6) ? (LenVect_[0]+1):
            (off < 6) ? 1:0;

      if (Dim_ >= 3 && max > 0)
        if (max < 2)
          max = 2;

      max *= sizeof(int*);
      Vect2_ = (off == 6) ?
      (int**)RawAllocateWith(MEMMATRIX, max):
      (off < 6) ? (Vect3_[0] = (int**)RawAllocateWith(MEMMATRIX, max)):
      NULL;
      
      Array_ = (off == 0) ? (void*)Vect8_:
               (off == 1) ? (void*)Vect7_:
               (off == 2) ? (void*)Vect6_:
               (off == 3) ? (void*)Vect5_:
               (off == 4) ? (void*)Vect4_:
               (off == 5) ? (void*)Vect3_:
               (off == 6) ? (void*)Vect2_:NULL;

      if (Array_)
      {
        Vect1_ = Vect2_[0] = (int*)RawAllocateWith(MEMMATRIX, sizeof(int) * sz);
        if (Dim_ >= 3)
          Vect2_[1] = (int*)RawAllocateWith(MEMMATRIX, sizeof(int) * 2);

        // Setting vector size and max
        Vect1_[0] = LenVect_[0];
        Vect1_[1] = 0;

        // Setting dealloc/sharedmem flags
        memset(Vect1_+1, 0, sizeof(int) * (sz - 1));
        InitMatrix(Array_, Dim_+off, off, Index_+off, LenVect_);
      }
    }
    else
    {
      if (Dim_ == 2)
      {
        if (InRange(1, PosVect_[1], SizeOfVector(Dim_, 0, LenVect_) - 1))
          pos = PosVect_[1];
        else
          pos = 0;

        if (pos)
        {
          sz = SizeOfVector(Dim_, 1, LenVect_);
          if (sz < 2)
            return NULL;

          Vect2_ = (int**)Array_;
          Contain_ = Vect2_[0];
          
          Vect1_ = Vect2_[pos] =
          (int*)RawAllocateWith(MEMMATRIX, sizeof(int) * sz);

          // Setting vector size and max
          Vect1_[0] = LenVect_[1];
          Error_ = false;
        }
        else
        {
          Error_ = true;
          return NULL;
        }
      }
      else
      {
        Vect8_ = (off == 0) ? (int********)Array_:NULL;
        Vect7_ = (off == 1) ? (int*******)Array_:NULL;
        Vect6_ = (off == 2) ? (int******)Array_:NULL;
        Vect5_ = (off == 3) ? (int*****)Array_:NULL;
        Vect4_ = (off == 4) ? (int****)Array_:NULL;
        Vect3_ = (off == 5) ? (int***)Array_:NULL;
        Vect2_ = (off == 6) ? (int**)Array_:NULL;
      
        RefMatrixDone_ = false;
        sz = SizeOfVector(Dim_, Index_, LenVect_);
        if (sz < 2)
          return NULL;

        xDim_ = Dim_;
        Dim_ += off;
        xIndex_ = Index_;
        Index_ += off;

        for (x = 1; x < xDim_; x++)
        {
          xOff_ = x + off;

          if ((Error_ = (x <= xIndex_)) &&
              InRange(x, PosVect_[x], SizeOfVector(xDim_, x-1, LenVect_) - 1))
            pos = PosVect_[x];
          else
            pos = 0;

          if (!pos && Error_)
            return NULL;
          else
            Error_ = false;
                        
          switch (xOff_)
          {
            case 1:
              if (pos)
              {
                if (xOff_ == Index_ && !RefMatrixDone_)
                {
                  if (Vect8_[pos])
                    Vect7_ = Vect8_[pos];
                  else
                  {
                    Contain2_ = Vect8_[0][0][0][0][0][0];
                    Contain_ = Contain2_[0];
                    Vect7_ = Vect8_[pos] =
                    (int*******)RawAllocateWith(MEMMATRIX, sizeof(int******) * sz);

                    IncCountOfVector(Dim_, Index_, off, Vect1_, Contain_, Contain2_);
                    if (CountOfVector(Dim_, Index_, off, Vect1_, Contain_, Contain2_) > Contain_[0])
                      ++Contain_[0];

                    InitMatrix(Vect7_, Dim_, off, Index_, LenVect_);
                    SetRefArrays(Vect7_, Dim_, off, Index_, LenVect_, PosVect_, Error_);
                    RefMatrixDone_ = true;
                  }
                }
                else
                {
                  if (Vect8_[pos])
                    Vect7_ = Vect8_[pos];
                  else
                  {
                    Contain2_ = Vect8_[0][0][0][0][0][0];
                    Contain_ = Contain2_[0];
                    Vect7_ = Vect8_[pos] =
                    (int*******)RawAllocateWith(MEMMATRIX, sizeof(int******));

                    IncCountOfVector(Dim_, Index_, off, Vect1_, Contain_, Contain2_);
                    if (CountOfVector(Dim_, Index_, off, Vect1_, Contain_, Contain2_) > Contain_[0])
                      ++Contain_[0];
                  }
                }
              }
              break;

            case 2:
              if (pos)
              {
                if (xOff_ == Index_ && !RefMatrixDone_)
                {
                  if (Vect7_[pos])
                    Vect6_ = Vect7_[pos];
                  else
                  {
                    Contain2_ = Vect7_[0][0][0][0][0];
                    Contain_ = Contain2_[0];
                    Vect6_ = Vect7_[pos] =
                    (int******)RawAllocateWith(MEMMATRIX, sizeof(int*****) * sz);

                    IncCountOfVector(Dim_, Index_, off, Vect1_, Contain_, Contain2_);
                    if (CountOfVector(Dim_, Index_, off, Vect1_, Contain_, Contain2_) > Contain_[0])
                      ++Contain_[0];

                    InitMatrix(Vect6_, Dim_, off, Index_, LenVect_);
                    SetRefArrays(Vect6_, Dim_, off, Index_, LenVect_, PosVect_, Error_);
                    RefMatrixDone_ = true;
                  }
                }
                else
                {
                  if (Vect7_[pos])
                    Vect6_ = Vect7_[pos];
                  else
                  {
                    Contain2_ = Vect7_[0][0][0][0][0];
                    Contain_ = Contain2_[0];
                    Vect6_ = Vect7_[pos] =
                    (int******)RawAllocateWith(MEMMATRIX, sizeof(int*****));

                    IncCountOfVector(Dim_, Index_, off, Vect1_, Contain_, Contain2_);
                    if (CountOfVector(Dim_, Index_, off, Vect1_, Contain_, Contain2_) > Contain_[0])
                      ++Contain_[0];
                  }
                }
              }
              else if (x != 1)
              {
                if (Vect7_[0])
                  Vect6_ = Vect7_[0];
                else
                  Vect6_ = Vect7_[0] =
                  (int******)RawAllocateWith(MEMMATRIX, sizeof(int*****));
              }              
              break;
              
            case 3:
              if (pos)
              {
                if (xOff_ == Index_ && !RefMatrixDone_)
                {
                  if (Vect6_[pos])
                    Vect5_ = Vect6_[pos];                  
                  else
                  {
                    Contain2_ = Vect6_[0][0][0][0];
                    Contain_ = Contain2_[0];
                    Vect5_ = Vect6_[pos] =
                    (int*****)RawAllocateWith(MEMMATRIX, sizeof(int****) * sz);

                    IncCountOfVector(Dim_, Index_, off, Vect1_, Contain_, Contain2_);
                    if (CountOfVector(Dim_, Index_, off, Vect1_, Contain_, Contain2_) > Contain_[0])
                      ++Contain_[0];

                    InitMatrix(Vect5_, Dim_, off, Index_, LenVect_);
                    SetRefArrays(Vect5_, Dim_, off, Index_, LenVect_, PosVect_, Error_);
                    RefMatrixDone_ = true;
                  }
                }
                else
                {
                  if (Vect6_[pos])
                    Vect5_ = Vect6_[pos];
                  else
                  {
                    Contain2_ = Vect6_[0][0][0][0];
                    Contain_ = Contain2_[0];
                    Vect5_ = Vect6_[pos] =
                    (int*****)RawAllocateWith(MEMMATRIX, sizeof(int****));

                    IncCountOfVector(Dim_, Index_, off, Vect1_, Contain_, Contain2_);
                    if (CountOfVector(Dim_, Index_, off, Vect1_, Contain_, Contain2_) > Contain_[0])
                      ++Contain_[0];                      
                  }
                }
              }
              else if (x != 1)
              {
                if (Vect6_[0])
                  Vect5_ = Vect6_[0];
                else
                  Vect5_ = Vect6_[0] =
                  (int*****)RawAllocateWith(MEMMATRIX, sizeof(int****));
              }              
              break;
            
            case 4:
              if (pos)
              {
                if (xOff_ == Index_ && !RefMatrixDone_)
                {
                  if (Vect5_[pos])
                    Vect4_ = Vect5_[pos];
                  else
                  {
                    Contain2_ = Vect5_[0][0][0];
                    Contain_ = Contain2_[0];
                    Vect4_ = Vect5_[pos] =
                    (int****)RawAllocateWith(MEMMATRIX, sizeof(int***) * sz);

                    IncCountOfVector(Dim_, Index_, off, Vect1_, Contain_, Contain2_);
                    if (CountOfVector(Dim_, Index_, off, Vect1_, Contain_, Contain2_) > Contain_[0])
                      ++Contain_[0];

                    InitMatrix(Vect4_, Dim_, off, Index_, LenVect_);
                    SetRefArrays(Vect4_, Dim_, off, Index_, LenVect_, PosVect_, Error_);
                    RefMatrixDone_ = true;
                  }
                }
                else
                {
                  if (Vect5_[pos])
                    Vect4_ = Vect5_[pos];
                  else
                  {
                    Contain2_ = Vect5_[0][0][0];
                    Contain_ = Contain2_[0];
                    Vect4_ = Vect5_[pos] =
                    (int****)RawAllocateWith(MEMMATRIX, sizeof(int***));

                    IncCountOfVector(Dim_, Index_, off, Vect1_, Contain_, Contain2_);
                    if (CountOfVector(Dim_, Index_, off, Vect1_, Contain_, Contain2_) > Contain_[0])
                      ++Contain_[0];
                  }
                }
              }
              else if (x != 1)
              {
                if (Vect5_[0])
                  Vect4_ = Vect5_[0];
                else
                  Vect4_ = Vect5_[0] =
                  (int****)RawAllocateWith(MEMMATRIX, sizeof(int***));
              }
              break;
            
            case 5:
              if (pos)
              {
                if (xOff_ == Index_ && !RefMatrixDone_)
                {
                  if (Vect4_[pos])
                    Vect3_ = Vect4_[pos];
                  else
                  {
                    Contain2_ = Vect4_[0][0];
                    Contain_ = Contain2_[0];
                    Vect3_ = Vect4_[pos] =
                    (int***)RawAllocateWith(MEMMATRIX, sizeof(int**) * sz);

                    IncCountOfVector(Dim_, Index_, off, Vect1_, Contain_, Contain2_);
                    if (CountOfVector(Dim_, Index_, off, Vect1_, Contain_, Contain2_) > Contain_[0])
                      ++Contain_[0];

                    InitMatrix(Vect3_, Dim_, off, Index_, LenVect_);
                    SetRefArrays(Vect3_, Dim_, off, Index_, LenVect_, PosVect_, Error_);
                    RefMatrixDone_ = true;
                  }
                }
                else
                {
                  if (Vect4_[pos])
                    Vect3_ = Vect4_[pos];
                  else
                  {
                    Contain2_ = Vect4_[0][0];
                    Contain_ = Contain2_[0];
                    Vect3_ = Vect4_[pos] =
                    (int***)RawAllocateWith(MEMMATRIX, sizeof(int**));

                    IncCountOfVector(Dim_, Index_, off, Vect1_, Contain_, Contain2_);
                    if (CountOfVector(Dim_, Index_, off, Vect1_, Contain_, Contain2_) > Contain_[0])
                      ++Contain_[0];
                  }
                }
              }
              else if (x != 1)
              {
                if (Vect4_[0])
                  Vect3_ = Vect4_[0];
                else
                  Vect3_ = Vect4_[0] =
                  (int***)RawAllocateWith(MEMMATRIX, sizeof(int**));
              }
              break;

            case 6:
              if (pos)
              {
                if (xOff_ == Index_ && !RefMatrixDone_)
                {
                  if (Vect3_[pos])
                    Vect2_ = Vect3_[pos];
                  else
                  {
                    Contain2_ = Vect3_[0];
                    Contain_ = Contain2_[0];
                    Vect2_ = Vect3_[pos] =
                    (int**)RawAllocateWith(MEMMATRIX, sizeof(int*) * sz);

                    IncCountOfVector(Dim_, Index_, off, Vect1_, Contain_, Contain2_);
                    if (CountOfVector(Dim_, Index_, off, Vect1_, Contain_, Contain2_) > Contain_[0])
                      ++Contain_[0];

                    InitMatrix(Vect2_, Dim_, off, Index_, LenVect_);
                    SetRefArrays(Vect2_, Dim_, off, Index_, LenVect_, PosVect_, Error_);
                    RefMatrixDone_ = true;
                  }
                }
                else
                {
                  if (Vect3_[pos])
                    Vect2_ = Vect3_[pos];
                  else
                  {
                    Contain2_ = Vect3_[0];
                    Contain_ = Contain2_[0];
                    Vect2_ = Vect3_[pos] =
                    (int**)RawAllocateWith(MEMMATRIX, sizeof(int*));

                    IncCountOfVector(Dim_, Index_, off, Vect1_, Contain_, Contain2_);
                    if (CountOfVector(Dim_, Index_, off, Vect1_, Contain_, Contain2_) > Contain_[0])
                      ++Contain_[0];
                  }
                }
              }
              else if (x != 1)              
              {
                if (Vect3_[0])
                  Vect2_ = Vect3_[0];
                else
                  Vect2_ = Vect3_[0] =
                  (int**)RawAllocateWith(MEMMATRIX, sizeof(int*));
              }
              break;

            case 7:
              if (pos)
              {
                if (xOff_ == Index_ && !RefMatrixDone_)
                {
                  if (Vect2_[pos])
                    Vect1_ = Vect2_[pos];
                  else
                  {
                    Contain2_ = Vect2_;
                    Contain_ = Vect2_[0];
                    Vect1_ = Vect2_[pos] =
                    (int*)RawAllocateWith(MEMMATRIX, sizeof(int) * sz);
                    Vect1_[0] = Vect1_[1] = 0;

                    IncCountOfVector(Dim_, Index_, off, Vect1_, Contain_, Contain2_);
                    if (CountOfVector(Dim_, Index_, off, Vect1_, Contain_, Contain2_) > Contain_[0])
                      ++Contain_[0];

                    InitMatrix(Vect1_, Dim_, off, Index_, LenVect_);
                    RefMatrixDone_ = true;
                  }
                }
                else
                {
                  if (Vect2_[pos])
                    Vect1_ = Vect2_[pos];
                  else
                  {
                    Contain2_ = Vect2_;
                    Contain_ = Vect2_[0];
                    Vect1_ = Vect2_[pos] =
                    (int*)RawAllocateWith(MEMMATRIX, sizeof(int) * 2);
                    Vect1_[0] = Vect1_[1] = 0;

                    IncCountOfVector(Dim_, Index_, off, Vect1_, Contain_, Contain2_);
                    if (CountOfVector(Dim_, Index_, off, Vect1_, Contain_, Contain2_) > Contain_[0])
                      ++Contain_[0];
                  }
                }
              }
              else if (x != 1)
              {
                if (Vect2_[0])
                  Vect1_ = Vect2_[0];
                else
                {
                  Vect1_ = Vect2_[0] =
                  (int*)RawAllocateWith(MEMMATRIX, sizeof(int) * sz);
                  Vect1_[0] = Vect1_[1] = 0;
                }
              }
          }
        }

        // Setting vector size
        Vect1_[0] = LenVect_[xIndex_];

        // Setting dealloc/sharedmem flags
        if (xIndex_ != xDim_ - 1)
          memset(Vect1_+1, 0, sizeof(int) * (sz - 1));
      }
    }
  }
  else if (Dim_ == 1)
  {
    // In calling InitArrayData function for 1D vector,
    // Index argument is assumed to hold the vector length.
    // LenVect argument is assumed to hold deallocated flag and
    // PosVect argument is assumed to hold shallowcopy flag.
    
    sz = Index_;
    Vect1_ = (int*)RawAllocateWith(MEMMATRIX, sizeof(int) * sz);
    Vect1_[0] = sz;
    Array_ = Vect1_;

    if (LenVect_)
      *LenVect_ = 0;

    if (PosVect_)
      *PosVect_ = 0;

    Error_ = false;
  }

  return Array_;
}

/****************************************************************************/
NestedArray3::NestedArray3(int* Lengths_):
_Array(NULL),
_MatrixIndex(0),
_LenVect(Lengths_ ?
           (int*)memmove(RawAllocateWith(MEMMATRIX, sizeof(int) * SIZE), Lengths_, sizeof(int) * SIZE):
           NULL),
_PosVect((int*)memset(RawAllocateWith(MEMMATRIX, sizeof(int) * (SIZE+1)), 0, sizeof(int) * (SIZE+1))),
_Error(false),
_Shared(0),
_Size(0),
_OffSets((int*)memset(RawAllocateWith(MEMMATRIX, sizeof(int) * (SIZE*2)), 0, sizeof(int) * (SIZE*2))),
_OffSetsLen(SIZE*2)
{
  int x;
  for (x = 1; x < SIZE; x++)
    _PosVect[x] = x;

  _PosVect[0] = SIZE;
  _PosVect[SIZE] = 1;

  if (!Lengths_)
  {
    _LenVect = (int*)RawAllocateWith(MEMMATRIX, sizeof(int) * SIZE);
    _LenVect[0] = SIZE;
    
    for (x = 1; x < SIZE; x++)
      _LenVect[x] = 0;
  }
  
  InitArrayData(0);
}

/****************************************************************************/
NestedArray3::~NestedArray3()
{
  ::DeleteArray(_LenVect);
  ::DeleteArray(_PosVect);
  ::DeleteArray(_OffSets);
}

/****************************************************************************/
void* NestedArray3::InitArrayData(int Index_)
{
  ::InitArrayData(_Array, SIZE, Index_, _LenVect, _PosVect, _Error);
}

/****************************************************************************/
int NestedArray3::DeleteVector(int Index_)
{
}

/****************************************************************************/
int NestedArray3::IsDeallocated(int Index_)
{
}

/****************************************************************************/
int NestedArray3::ReplaceVector(int Index_, void* Replacement_, bool Replace_,
                                bool DeAlloc_)
{
}

/****************************************************************************/
int NestedArray3::RealVectorSize(int Index_)
{
}

/****************************************************************************/
int* NestedArray3::ParentReference(int Index_, int* ParentRefs_)
{
}

/****************************************************************************/
int* NestedArray3::LinearToCoords(int Pos_, int* OffSets_, int OffSetsLen_)
{
}

/****************************************************************************/
int NestedArray3::CoordsToLinear(int* OffSets_, int OffSetsLen_)
{
}

/****************************************************************************/
int* NestedArray3::CalculateOffsets()
{
}

/****************************************************************************/
int NestedArray3::ReadFromArray(int Index_)
{
}

/****************************************************************************/
void* NestedArray3::WriteToArray(int Index_, int Data_)
{
}

/****************************************************************************/
#if NESTARRAY_DEBUG
int main()
{

#if NESTARRAY_DEBUG1
#endif
}

#endif
/****************************************************************************/
#endif

