// Window line buffer ADT classes
// The window line buffer class define methods for maintaining a window line
// object as parts used in a window screen object in a text console. The
// window line buffer class also has methods for positioning procedures for
// setting the display position within the window screen, setting the default
// message and default input into the window line, showing, hiding, erasing
// and retrieval of the input data.
//
// The window line buffer class define methods for maintaining an array of
// window line buffers. Methods are defined to show, hide, erase and
// retrieve from the entire array of window line buffers assigned to the
// window screen object.
//
#ifndef WINLINE_H
#define WINLINE_H
/****************************************************************************/
#ifndef CTRLDEFS_H
  #include "ctrldefs.h"
#endif
#ifndef TEXTGRAPHICS_H
  #include "txtgraph.h"
#endif

#define WINLINE_DEBUG      0
#define WINLINE_DEBUGEXL   0  // test buffer at double input length

#define WINLINE_DEBUG1     0  // multi-line window buffers
#define WINLINE_DEBUG1A    0  // test line wrapping
#define WINLINE_DEBUG1B    0  // test in-window properties
#define WINLINE_DEBUG1C    0  // test multi-line adding of new lines

#define WINLINE_DEBUG2     0  // single line window buffer (extended)
#define WINLINE_DEBUG3     0  // Window mode functions debug

#define WINLINE_WSENDMOVE  1  // process ws for end of box cursor moves
#define WINLINE_ONLYCWSCHR 1  // only count white characters coded in
                              // standard ascii values when finding ws length

// WinLineBuffer operations:
//
// [winlinebuffer x] <-- [winlinebuffer] ... [winlinebuffer] <-- from start.
//
// pull characters from adjacent buffers and end buffer will pull characters
// from start buffer until max. wraparound length is reached which is:
//
//   ((total buffers) - 1) x Input Length
//
//   Input length is the same as buffer length or _maxbuf
//
// WinLineBuffer objects will be placed relative to the windows bounded box
// as specified by WindowsLine::winboxmaxx and WindowsLine::winboxmaxy
//
// Any winlinebuffer objects with size and starting positions surpassing
// the winbox dimensions will be hidden and not shown.
//
// The WinLineBuffer InFocusMonitor is disabled for this class's purpose
//
// The right arrow key would perform the chain wraparound as described above
// until the wraparound index reaches the wraparound length limit.
//
// The left arrow key would reverse the chain wraparound until the
// wraparound index reaches zero.
//
// win:LEFT, linux:KEY_LEFT:
// 1.0 1.1xxxxx1.2xxxxx1.3_____1.4
// 1.1 stay at 0 pos
// 1.2 move to previous char
// 1.3 move to previous char
// 1.4 move back one char in refline object
// 
// win:RIGHT, linux:KEY_RIGHT:
// 2.0 2.1xxxxx2.2xxxxx2.3_____2.4
// 2.1 move to next char
// 2.2 move to next char
// 2.3 move forward one char in refline object
// 2.4 move forward one char in refline object
// 
// win:HOME, linux:KEY_HOME:
// 3.0 3.1xxxxx3.2xxxxx3.3_____3.4
// 3.1 stay at 0 pos
// 3.2 repeat win:LEFT or linux:KEY_LEFT until 0 pos reached
// 3.3 repeat win:LEFT or linux:KEY_LEFT until 0 pos reached
// 3.4 repeat move back one char in refline object until refline extindex==0,
//     repeat win:LEFT or linux:KEY_LEFT unto 0 pos reached
// 
// win:END, linux:KEY_END:
// 4.0 4.1xxxxx4.2xxxxx4.3_____4.4
// 4.1 repeat win:RIGHT or linux:KEY_RIGHT until
//     boxcnt==wrapmax AND index=endpt is reached
// 4.2 repeat win:RIGHT or linux:KEY_RIGHT until
//     boxcnt==wrapmax AND index=endpt is reached
// 4.3 stay at endpt pos
// 4.4 repeat move back one char in refline object until refline extindex==0,
//     stay at endpt pos
// 
// win:DELETE, linux:KEY_DC:
// 5.0 5.1xxxxx5.2xxxxx5.3_____5.4
// 5.1 delete to next char and chainpull all chars back one position
// 5.2 delete to next char and chainpull all chars back one position
// 5.3 shift up next line and chainpush to current line adding more box chains
//     as required to store contents
// 5.4 shift up next line and chainpush to current line adding more box chains
//     as required to store contents
//     chainfill blanks from endpt pos at 5.3 to start of shifted up line at 5.4
// 
// win:ENTER, linux:10:
// 6.0 6.1xxxxx6.2xxxxx6.3_____6.4
// 6.1 shift down whole line buffer of current line and null terminate at 0 pos
// 6.2 shift down line buffer at 6.2 of current line and null terminate at 6.2
// 6.3 create next line with 0 length and null terminate at endpt pos trim ws
// 6.4 create next line with 0 length and null terminate at endpt pos trim ws
// 
// win:BACK, linux:KEY_BACKSPACE:
// 7.0 7.1xxxxx7.2xxxxx7.3_____7.4
// 7.1 stay at 0 pos
// 7.2 move to previous char,
//     delete to next char and chainpull all chars back one position,
// 7.3 move to previous char,
//     delete to next char and chainpull all chars back one position,
// 7.4 move back one char in refline object
// 
// win:ESCAPE, linux:27:
// 8.0 8.1xxxxx8.2xxxxx8.3_____8.4
// 
// Deactivate from current WindowLine and WindowBlock and return to container
// widget object in all cases.
// 
// win:NEXT, linux:9
// 9.0 9.1xxxxx9.2xxxxx9.3_____9.4
// 
// If not in WindowMode then put in tab character if in window mode then
// move to next WindowLine object (same behavior as win:DOWN or linux:KEY_DOWN)
// 
// win:UP, linux:KEY_UP:
// 10.0 10.1xxxxx10.2xxxxx10.3_____10.4
// 
// If in window mode then move to previous WindowLine object in all cases
// All internal values of refline object is adjusted so previous line will be
// at same column as the current line.
// 
// win:DOWN, linux:KEY_DOWN:
// 11.0 11.1xxxxx11.2xxxxx11.3_____11.4
// 
// If in window mode then move to next WindowLine object in all cases
// All internal values of refline object is adjusted so next line will be
// at same column as the current line.
//
/****************************************************************************/
class ControlWindow;
class WindowRefLineInfo;
class WindowLineBuffer;
class WindowLine;

class ChainPosInfo
{
  protected:
    void CopyInfoToThis(WindowLineBuffer* Ptr_);
    void CopyToPosInfo(WindowLineBuffer* Ptr_);

  public:
    bool _enterb;
    bool _vbuf;
    bool _inwbox;
    bool _chainable;
    bool _error;
    int _vmax;
    int _vendpt;

    int _WrapMax;
    int _WrapIndex;

    int _reallen;
    int _copylen;
    int _index;
    int _startpt;
    int _endpt;
    int _boxpt;

    int _ChainPos;         // ChainPos: 0=vbufstart, 1=vbufend, 2=inwbox
    int _backshift;
    int _boxendterm;
    int _refshendterm;
    int _reflineshift;

    int _curbufpos;
    int _endptdex;
    int _refshindex;
    int _refshendpt;

    int _endstrdex;
    int _refshendstr;

    char* _baseposbuf;
    char* _refshbuf;
    char* _refshjbuf;

    WindowLineBuffer* _BufInfo;
    ChainPosInfo* _StartInfo;    // Starting Info at real part of vbuf str
    ChainPosInfo* _EndInfo;      // Ending Info at virtual part of vbuf str

    ChainPosInfo();
    ~ChainPosInfo();

    void SetupInfoPtrs(WindowLineBuffer* Ptr_, int SeqPos_);
    void SetPushInfo(int start);
    void SetPullInfo(int start);

    void RestorePrevInfo();
    void ReturnToThis();

    int GetRefLineShift(bool MatchEndPts_=true);

    char* GetBasePosBuffer();
    char* GetRefshBuffer();
    char* GetRefshJoinBuf();

    int GetBufBaseIndexPos(bool ReCalc_=false);
    int GetBufBaseEndPt(bool ReCalc_=false);
    int GetRefshIndexPos(bool ReCalc_=false);
    int GetRefshEndPt(bool ReCalc_=false);
    int GetRefshCursorPos(int Base_, int CurPos_, bool ReCalc_=false);

    int GetBufBaseEndStr(bool ReCalc_=false, bool MatchEndPts_=true);
    int GetRefshEndStr(bool ReCalc_=false, bool MatchEndPts_=true);
};

class WindowLineBufferType : public TextControl
{
  public:
    virtual void SetInputStringAtPos(char* str, int len, int pos, bool endterm_) = 0;
    virtual void FillCharAtPos(char ch, int pos, bool endterm_) = 0;

    virtual void ChainSetOwner(ControlWindow* Pwin_, TextControl* TxtCtrl_) = 0;
    virtual void CreateReturnBuffer(int Size_) = 0;
    virtual void SetReturnBuffer(WindowLineBufferType* Lbuf_,
                                 char* RetBufPtr_, bool Alloc_, int Size_=0) = 0;
    virtual void SetBufferChainNum(WindowLineBufferType* Lbuf_, int Index_) = 0;
    virtual void SetWindowLineNum(int Index_) = 0;
    virtual int InputLength() const = 0;

    virtual int IsFullBox() const = 0;
    virtual int RestrictFullBox() const = 0;
    virtual bool EndBrksActive() const = 0;
    virtual bool ShouldTrimWs() const = 0;
    virtual int BufferIndex() const = 0;
    virtual int InBoxOffset() const = 0;
    virtual int InWBoxCount() const = 0;
};

class WindowLineBuffer : public WindowLineBufferType
{
  friend const int* InsertDetect(void* Owner_, WindowLineBuffer*, const int* Ptr_);
  friend class WindowRefLineInfo;
  friend class WindowLine;
  friend class ChainPosInfo;

  protected:
    enum { BLANK        = ' ',
           UNINIT_NL    = -1,
           FILLED_NL    = 0,
           INIT_NL      = 1,

           INPUT_BUFTYPE  = 0x01,
           RETURN_BUFTYPE = 0x02,
           JOIN_BUFTYPE   = 0x04,
           TRANS_BUFTYPE  = 0x08,
           
           SPACE_CHAR   = 0xAE,
           TAB_CHAR     = 0xAF,
           RETURN_CHAR  = 0xB0,
           NEWLINE_CHAR = 0xB1,
           NOJOIN_CHAR  = 0xB2,
           MAXBUFFER    = 128 };
#if (defined(__TURBOC__) | defined(__BORLANDC__) | defined(__DJGPP__))
    enum { CONTMORE_RIGHT = 0x10,
           CONTMORE_LEFT = 0x11 };
    enum { CONTMORE_HFG = LIGHTGREEN,
           CONTMORE_HBG = BLUE };
#elif (defined(__linux__) | defined(__unix__))
    enum { CONTMORE_RIGHT = '>',
           CONTMORE_LEFT = '<' };
    enum { CONTMORE_HFG = PINK,
           CONTMORE_HBG = GREEN };
#endif

    struct SharedChainData
    {
      WindowLine* _owner;
      WindowLineBuffer* _parent;
      mutable const WindowLineBuffer* _reentryptr;

      char* _nullcharstr;    // null character return string
      int _padcnt;           // character pad counter
      int _padmax;           // maximum character pad limit
      int _cpadpos;          // found eol index pos in CharPadToEnd method
      int _cpadchain;        // BufChainNum at eol point in CharPadToEnd method
      bool _cpadset;         // _cpadpos and _cpadchain is set in CharPadToEnd method

      bool _InWinMode;       // in text editor window mode
      bool _PrevLineReady;   // previous text line available in text editor window
      bool _NextLineReady;   // next text line available in text editor window
      bool _countjoinedch;   // count joined characters in NextLength method
      bool _movedone;        // key move done for line buffer
      bool _BoxPosSaved;     // indicate box position data saved
      bool _BoxPosTrans;     // indicate box position data transferred
      bool _EndPtRefShift;   // End point reference shift

      int _SavedOffset;      // Saved inputbox offset or count
      int _SavedIndex;       // Saved input buffer index
      int _SavedExtIndex;    // Saved extended index of reference line
      int _SavedBoxDiff;     // Saved box pos diff from box index value
      
      int _NextOffset;       // Transferred inputbox offset or count
      int _NextIndex;        // Transferred input buffer index
      int _NextExtIndex;     // Transferred extended index of reference line
      int _NextBoxDiff;      // Transferred box pos diff from box index value
      
      int _numnextlines;     // number of nextline buffers
      int _WinLineNum;       // Window line number

      SharedChainData(WindowLineBuffer* wlbptr=NULL);
      ~SharedChainData();

      void SetOwner(WindowLine* wlptr);
      void SetParent(WindowLineBuffer* wlbptr);
    };

    ColorInfo _ColorInfo;
    ControlWindow* _Owner;
    WindowLineBuffer* _nextline;
    WindowRefLineInfo* _refline;
    ChainPosInfo* _ChainPosPtr;

    bool _SharedDataErased;
    SharedChainData* _SharedData;

    char _brackets[2];
    Boolean _Enterable;    // Control could be entered into
    Boolean _Selectable;   // Indicate whether control can be
                           // selected like a button or checkbox
    char* _msg;            // prompt message buffer
    char* _buffer;         // input buffer
    char* _retbuf;         // return buffer
    char* _joinbuf;        // buffer containing char. buffer joined status
    char* _tranbuf;        // container buffer for extended ascii chars.
    bool _bufferalloc;     // input buffer allocated
    bool _retbufalloc;     // memory allocated to return buffer
    bool _joinbufalloc;    // join status buffer allocated
    bool _tranbufalloc;    // extended ascii buffer allocated
    int _bufferscleared;   // buffer types cleared in buffer chain

    int _bufferoffset;     // virtual buffer memory offset
    int _inboxoffset;      // sliding inputbox offset into virtual buffer
    int _inwboxcnt;        // wraparound boxex wrap count
    
    bool _MovedToEnd;      // cursor moved to end of text box chain
    int _EndCharInserted;  // latest end point char. inserted into text box
    bool _EndCharReverted; // end point char. reverted back into position
    bool _DelKeyDone;      // last key pressed by user was char delete key
    int _DelWrapOffset;    // Difference in WrapMax from box offset value
                           // resulting from char delete key being pressed

    int _linestatus;       // line status : { UNINIT, JOINED, INIT }
    int _maxbuf;           // maximum size of input buffer
                           // not including NULL terminator
    int _key;              // input key
    int _extkey;           // extended input key
    int _extended;         // extended key indicator
    int _index;            // input buffer index
    int _len;              // maximum input length
    int _retbufsize;       // return buffer size

    int _minlen;           // minimum input length
    int _chainminlen;      // remaindered min length for chained box
    int _xpos;             // horizontal coordinate position
    int _ypos;             // vertical coordinate position
    int _noempty;          // no input disallow flag
    int _InsertMode;       // insert mode
    int _Escaped;          // prompt box Escaped
    int _Hidden;           // Text dialog hidden flag
    int _TextWidgetActive; // Active flag for this text widget
    int _ChainPos;         // The position within the chain
    int _InFocus;          // Text widget in focus flag

    int _SavedCursX;       // Saved cursor x position
    int _SavedCursY;       // Saved cursor y position

    int _BufChainNum;      // Buffer chain number
    bool _WrapChars;       // Wrap chars from next buffer
    
    bool _RestFullBox;     // Further input into full box is disallowed
    bool _EndBrksActive;   // Ending brackets active at left-right ends
    bool _TrimWs;          // Trim white space from input line (default: false)

    // Internal data monitoring function pointers
    const int*(*_insert_monitor)(void*, WindowLineBuffer*, const int*);    // internal use
    const char*(*_buffer_monitor)(void*, WindowLineBuffer*, const char*);
    const int*(*_key_monitor)(void*, WindowLineBuffer*, const int*);
    const int*(*_ext_monitor)(void*, WindowLineBuffer*, const int*);
    const int*(*_index_monitor)(void*, WindowLineBuffer*, const int*);
    const int*(*_escaped_monitor)(void*, WindowLineBuffer*, const int*);
    const int*(*_hidden_monitor)(void*, WindowLineBuffer*, const int*);
    const int*(*_active_monitor)(void*, WindowLineBuffer*, const int*);
    const int*(*_infocus_monitor)(void*, WindowLineBuffer*, const int*);
    const int*(*_inboxoffset_monitor)(void*, WindowLineBuffer*, const int*);
    const int*(*_winlinenum_monitor)(void*, WindowLineBuffer*, const int*);

    // Handles to external objects, used by data monitors
    void* _insert_handle;
    void* _buffer_handle;
    void* _key_handle;
    void* _ext_handle;
    void* _index_handle;
    void* _escaped_handle;
    void* _hidden_handle;
    void* _active_handle;
    void* _infocus_handle;
    void* _inboxoffset_handle;
    void* _winlinenum_handle;

    // Monitor functions execution and management methods
    const char* cMonitor(const char*(*monptr_)(void*, WindowLineBuffer*, const char*), void* This_, const char* ptr_);
    const int* iMonitor(const int*(*monptr_)(void*, WindowLineBuffer*, const int*), void* This_, const int* ptr_);
    void SaveCursorPosition();
    void RestoreCursorPosition();

    const int* MonitorInsertMode(void* This_, const int*(*FnPtr_)(void*, WindowLineBuffer*, const int*));

    // Promptbox chaining internal methods
    void SetChainPos(int Pos_);

    // Insert state methods
    void NotifyInsertMode(const int* ModePtr_);
    void SetInsertMode(int Mode_, WindowLineBuffer* Parent_=NULL);
    inline int GetInsertMode()
        { return _InsertMode; }

    // show prompt box methods
    virtual void ShowMessage();
    virtual void ShowInput(int status_);
    virtual void ResetPrompt();

    // Clear prompt box method
    void PrintBlanks(int Max_);

    // input validation / testing methods
    int* WordChr(int* src_, int val_, size_t n);
    int IsValidString(char* str);
    int IsValidLength() const;
    int IsUserKey(int key);
    int ShouldEnd(int key, int zerotozero);
    int IsValidCtrlKey(int key);

    // console input retrieval methods
    int GetKey();
    inline int GetExtKey()
        { return _extkey; }
    int GetLine(int status_);
    int ReadInput(int status_);
    void ResetInputState(int status_);
    void SetEnterable(Boolean Flag_);
    void SetSelectable(Boolean Flag_);
#if !(defined(__TURBOC__) | defined(__BORLANDC__) | defined(__DJGPP__)) & (defined(__linux__) | defined(__unix__))
    int TranslateKeys(int key);
#endif

    // keyboard character output methods
    bool PutExtKey(int key);
    bool PutStdKey(int key);
    void SetMoveDone(bool move_);
    bool MoveDone() const;
    void SetRefLineMoveDone(bool move_);
    bool RefLineMoveDone() const;
    void CommitRefLinePendingMove();
    void SetRefLineMovePending(bool pending_, bool done_);
    bool RefLineMovePending() const;

    // String push/pull methods
    const char* RawStrBuffer(int index_) const;
    char* RawStrBuffer(int index_, bool Revert_=true, int* shift_=NULL);    
    int RawStrBufLen(int index_) const;
    int RawStrBufLen(int index_, int* shift_=NULL);
    
    bool IsEquivSpace(char ch) const;
    Uchar Convert(char ch, bool* Converted_=NULL) const;
    char Revert(Uchar ch, bool* Reverted_=NULL) const;
    char* Convert(char* Str_);
    char* Revert(char* Str_);
    char* StrRetrieve(bool ExtRefLine_, bool Revert_, int BackShift_=0);

    static char* RemoveTrailSpaceChars(char* Str_, bool UnReverted_);
    static bool InChSet(char ch, const char* chset, size_t* Index_=NULL);
    static void FillString(char* Str_, char FillCh_, int Start_, int End_);
    bool CursorMoveCondMet(int WrapIndex_, int WrapMax_, int endpt_, int NonWsLen_);
    void ShiftDownAndSetStr(char* NewStr_);
    int HasNonNullChars(const char* str_) const;
    int HasNonWsChars(const char* str_) const;
    int NextLengthHelper() const;
    int TakeNextLastCharHelper(bool Peek_=false);    
    int GetRefLineShift();
    int CalcBackShift(const char* retstr_);
    void ChainPush(char ch, char* join_, ChainPosInfo* Info_);
    void ChainJoin(char ch);
    char ChainPull(char* join_, char* chain_, ChainPosInfo* Info_);
    bool Push(int start, char ch, char* join_, ChainPosInfo* Info_);
    void Join(char ch);
    char Pull(int start, char* join_, char* chain_, ChainPosInfo* Info_);
    char PeekNextChar(int Index_=0);
    void DetermineWrapChars();
    void ChainRemoveRetBuf(char* RetBufPtr_);
    bool ChainedBufferEndReached(int* UnJoinedLen_=NULL) const;
    int VbufStrLenNoTerm(bool NoTerm_=true) const;
    int VbufStrLen(int* lastnwspos_=NULL, int* StartPos_=NULL,
                   int* EndPos_=NULL, char* LastCharp_=NULL) const;

    // Extended box methods
    int CursorAtEOL(bool Reversed_);
    bool CompareVal(int Val1_, int Val2_, int Relation_);
    char* NewCString(int len_, char pad_, bool AddZeroTerm_);
    inline bool BufferTypeCleared(int BufType_)
        { return (_bufferscleared & BufType_); }
    bool BuffersCleared() const;
    int ClearUnusedBuffers();
    int RestoreUnusedBuffers();
    void SetBufferTypeCleared(int BufType_, bool Cleared_);
    bool EndBoxEmpty();
    bool AllBoxFilled();
    void ResetBoxData();
    void SetEscaped(bool Flag_);
    void ResetRefPos(int MoveToPos_);
    void ReAdjustLine(int Offset_);
    void ReAlignBoxIndex();
    void SetCountJoinedChars(bool Flag_);
    void EraseSharedChainData(WindowLineBuffer::SharedChainData* ptr);
    void CharPadToEnd(char ch, WindowLineBuffer::SharedChainData* ptr=NULL);
    bool IsAtEndPt(bool* PastEndPt_=NULL, WindowRefLineInfo* RefLinePtr_=NULL,
                   bool* RefLineMoved_=NULL, int* ExtIndex_=NULL, int* BoxDiff_=NULL);
    bool CreateExtendedAsciiBuffer();
    void ClearDelWrapOffset(int boxshift_, int WrapMax_);
    inline int GetMaxWrap(int WrapMax_) const
        { return (WrapMax_ + _DelWrapOffset); }
    bool AtMaxWrap(int WrapIndex_, int WrapMax_, int endpt_, int Relation_,
                   bool AtPosIndex_, bool InclDelWrapOff_,
                   bool* AtEnd_=NULL, bool* PastEnd_=NULL);
    bool ExtendedBoxMode(bool& vbuf_, bool& inwbox_,
                         int* EndReached_=NULL, int* NextNonWsLen_=NULL,
                         int* WrapIndex_=NULL, int* WrapMax_=NULL) const;

    inline WindowLineBuffer::SharedChainData* SharedData()
        { return ((!_SharedDataErased && _SharedData) ? _SharedData:NULL); }
    inline const WindowLineBuffer::SharedChainData* ConstSharedData() const
        { return ((!_SharedDataErased && _SharedData) ? _SharedData:NULL); }

    // Dynamic box chaining methods
    WindowLineBuffer* LastChainBox();
    WindowLineBuffer* IncrBoxChains(bool blankpad_);
    WindowLineBuffer* SetupChainObject(WindowLineBuffer* ptr, ControlWindow* Owner_);
    WindowLineBuffer* DetachNext(WindowLineBuffer* Chainp_, bool Setup_);
    WindowLineBuffer* DetachThis(WindowLineBuffer* Chainp_, bool Setup_);
    WindowLineBuffer* CopyEndLineCharData(WindowLine* owner_,
                                          WindowLineBuffer* ptr, bool Init_=false);
    WindowLineBuffer* InitFirstBox(WindowLine* owner_);
    WindowLineBuffer* AddBoxChain(WindowLineBuffer* next_, bool ChainPWinBox_);
    
    void ReCalcChainWbuffer(bool ReCalcSdp_, WindowLine* wlptr);
    char ConfirmBufferChains(WindowLineBuffer* root, int ChainNum_);
    bool RemoveBoxChains(WindowLineBuffer* wlptr);
    void SaveBoxPos();
    void RestoreBoxPos();
    void RestoreBoxPos(WindowLineBuffer* wlptr);

    // data member destruction method
    virtual void Clear();

    // promptbox constructors    
    WindowLineBuffer(char* msg, WindowLineBuffer* next_, WindowLine* owner_,
                     int Size_, int Length_=0);
    WindowLineBuffer(WindowLineBuffer& pb, int Size_, int Length_, int Offset_);
    WindowLineBuffer(WindowLineBuffer& pb, int Size_, int Length_=0);

    WindowLineBuffer* MakeVirtualLine(WindowLineBuffer* wpb);

    // Text color methods
    void SetupColors(int Mode_);
    void SetupColors(int Fg_, int Bg_);

    // status indicator return functions
    bool MovedToEnd() const
        { return _MovedToEnd; }
    bool EndCharInserted() const
        { return (_EndCharInserted && !_EndCharReverted); }
    bool EndCharReverted() const
        { return (_EndCharReverted && !_EndCharInserted); }
    bool DelKeyDone() const
        { return _DelKeyDone; }
    int DelWrapOffset() const
        { return _DelWrapOffset; }

  public:
    // WindowLineBuffer Destructor
    virtual ~WindowLineBuffer()
        { Clear(); }

    // Promptbox chaining method
    WindowLineBuffer* Chain(WindowLineBuffer* next_);
    inline void EraseChain()
        { _nextline = NULL; }
    inline WindowLineBuffer* NextChain()
        { return _nextline; }

    virtual int IsFullBox() const;
    virtual int RestrictFullBox() const;
    virtual bool EndBrksActive() const;
    virtual bool ShouldTrimWs() const;
    virtual int BufferIndex() const;
    virtual int InBoxOffset() const;
    virtual int InWBoxCount() const;

    void SetEndBrksActive(bool Flag_);
    void SetRestrictFullBox(bool Flag_);
    void SetTrimWs(bool Flag_);
    int EndBrksLength(bool OnGoto_) const;

    // virtual memory buffer methods
    virtual int VirtualBufferOffset(int* EndReached_=NULL,
                                    int* NextNonWsLen_=NULL,
                                    int* NextNonNullLen_=NULL) const;
    virtual int VirtualBufferMax() const;
    virtual int IsChainable(bool HasChars_) const;
    virtual int ThisLength() const;
    virtual int NextLength() const;
    virtual int TakeNextLastChar(bool Peek_=false);
    virtual int IsDummyLine() const;
    virtual int IsWindowLineBuffer() const;
    virtual int ChainPosition() const;
    virtual int BufferChainPosition() const;
    
    // promptbox assignment operator
    virtual WindowLineBuffer& operator = (WindowLineBuffer& pb);
    static WindowLineBuffer* Make(char* msg=NULL, WindowLineBuffer* next_=NULL,
                                  WindowLine* owner_=NULL, int Size_=0, int Length_=0);
    static WindowLineBuffer* Make(WindowLineBuffer& pbox, int Size_=0, int Length_=0);
    static WindowLineBuffer* MakeNextLine(WindowLineBuffer* wpb);

    // prompt box properties mutators
    virtual char* SetMessage(const char* str);
    virtual void RemoveMessage();
    virtual void BlankMessage();
    virtual void SetInputString(char* str);
    virtual void SetInputStringAtPos(char* str, int len, int pos, bool endterm_);
    virtual void FillCharAtPos(char ch, int pos, bool endterm_);
    virtual void PushInputString(char* str);
    virtual void SetInputLength(int len, bool MakeVbuf_);
    virtual int InputLength() const;
    virtual void DivBufSizeAsInputLength(int div_, bool MakeVbuf_);
    virtual void SetMinimumLength(int len);
    void SetNoEmpty(int flag=1);
    bool SetXYpos(int x, int y);
    bool SetXpos(int x);
    bool SetYpos(int y);
    int GetXpos() const;
    int GetYpos() const;
        
    // prompt box manipulators
    virtual void Show();
    virtual void ShowInput();
    virtual void Hide();
    virtual int Prompt(int status_);
    virtual void Erase();

    // Accelerator key methods
    virtual int GetHotKey() const;
    virtual int GetShortCut() const;
    virtual int GetHotKeyType() const;

    // text control selection/deselection mutators
    virtual void Select();
    virtual void Deselect();

    // Empty, valid and selected Text Control test
    virtual Boolean IsEnterable() const;
    virtual Boolean IsSelectable() const;
    virtual Boolean IsSelected() const;
    virtual Boolean IsEmpty() const;
    virtual Boolean IsEscaped() const;
    virtual Boolean IsValid() const;
    virtual Boolean IsActive() const;
    virtual Boolean IsInFocus() const;

    // inbox index methods
    virtual void DecrBoxIndex();
    virtual void IncrBoxIndex();
    virtual void ResetBoxIndex();
    virtual void SetBoxIndex(int Val_);

    // Activation/Deactivation for this text widget
    virtual void Activate();
    virtual void Deactivate();

    // input data retrieval method
    char LastKeyRead(int* Ext_=NULL) const;
    int LastExtKeyRead() const;
    char ChrRetrieve(int Index_);
    inline char* GetMessage()
        { return _msg; }
    inline char* StrRetrieve()
        { return StrRetrieve(false, true); }

    // Text control ownership methods
    virtual void ChainSetOwner(ControlWindow* Pwin_, TextControl* TxtCtrl_);
    virtual void SetOwner(ControlWindow* Pwin_, TextControl* TxtCtrl_);
    virtual Boolean IsOwner(ControlWindow* Pwin_) const;
    
    // Text color methods
    virtual void UseColor(Boolean Flag_=TRUE);
    virtual void SetNormalColors(int FgCol_, int BgCol_);
    virtual void SetHighLightColors(int FgCol_, int BgCol_);
    virtual void SetHotKeyColor(int Col_);
    virtual void SetBackGroundColor(int Col_);
    virtual void UseDefaultColors();

    virtual Boolean UsesColor() const;
    virtual int* NormalColors() const;
    virtual int* HighLightColors() const;    
    virtual int HotKeyColor() const;
    virtual int BackGroundColor() const;

    // WindowLineBufferType specific methods
    virtual void SetWindowLineNum(int Index_);
    virtual void SetBufferChainNum(WindowLineBufferType* Lbuf_, int Index_);
    virtual void CreateReturnBuffer(int Size_);
    virtual void SetReturnBuffer(WindowLineBufferType* Lbuf_,
                                 char* RetBufPtr_, bool Alloc_, int Size_=0);

    // Internal data monitoring methods
    const char* MonitorBuffer(void* This_, const char*(*FnPtr_)(void*, WindowLineBuffer*, const char*));
    const int* MonitorKey(void* This_, const int*(*FnPtr_)(void*, WindowLineBuffer*, const int*));
    const int* MonitorExtendedKey(void* This_, const int*(*FnPtr_)(void*, WindowLineBuffer*, const int*));
    const int* MonitorIndex(void* This_, const int*(*FnPtr_)(void*, WindowLineBuffer*, const int*));
    const int* MonitorEscapedStatus(void* This_, const int*(*FnPtr_)(void*, WindowLineBuffer*, const int*));
    const int* MonitorHiddenStatus(void* This_, const int*(*FnPtr_)(void*, WindowLineBuffer*, const int*));
    const int* MonitorActiveStatus(void* This_, const int*(*FnPtr_)(void*, WindowLineBuffer*, const int*));
    const int* MonitorInFocusStatus(void* This_, const int*(*FnPtr_)(void*, WindowLineBuffer*, const int*));
    const int* MonitorInboxOffset(void* This_, const int*(*FnPtr_)(void*, WindowLineBuffer*, const int*));
    const int* MonitorWinLineNum(void* This_, const int*(*FnPtr_)(void*, WindowLineBuffer*, const int*));
};

/****************************************************************************/
class WindowRefLineInfo : public WindowLineBufferType
{
  friend const int* InsertDetect(void* Owner_, WindowRefLineInfo*, const int* Ptr_);
  friend class WindowLine;
  friend class WindowLineBuffer;

  protected:
    enum { BLANK = ' ', MAXBUFFER = 128 };

    ColorInfo _ColorInfo;
    ControlWindow* _Owner;

    Boolean _Enterable;       // Control could be entered into
    Boolean _Selectable;      // Indicate whether control can be
                              // selected like a button or checkbox

    WindowLineBuffer* _parent;   // same as pbox[0]
    bool _movedone;        // key move done for dummy line buffer
    bool _haspendingmove;  // has pending move for when refline methods are
                           // not executed or replaced with other methods
    bool _pendingmove;     // pending move value to assign _movedone data
                           // member when pending move is committed
    bool _extindextoggled; // extended index have been toggled 0-->1 or 1-->0

    int _parentindex;   // same as _parent->_index
    int _inboxoffset;   // same as _parent->_inboxoffset
    int _inwboxcnt;     // same as _parent->_inwboxcnt
    bool _inwbox;       // same as _parent->inwbox_
    bool _vbuf;         // same as _parent->vbuf_

    int _maxbuf;        // maximum size of input buffer
                        // not including NULL terminator
    int _len;           // maximum input length
    int _minlen;        // minimum input length
    int _distance;      // absolute distance from zero position start pt.
    int _index;         // input buffer index
    int _extindex;      // extra char. lengths tacked onto end of line
    int _boxpos;        // actual cursor position relative to _len in box
    int _boxdiff;       // difference from cursor box pos. to index pos.

    int _key;              // input key
    int _extkey;           // extended input key
    int _extended;         // extended key indicator
    int _xpos;             // horizontal coordinate position
    int _ypos;             // vertical coordinate position
    int _noempty;          // no input disallow flag
    int _InsertMode;       // insert mode
    int _Escaped;          // prompt box Escaped
    int _Hidden;           // Text dialog hidden flag
    int _TextWidgetActive; // Active flag for this text widget
    int _InFocus;          // Text widget in focus flag

    int _SavedCursX;       // Saved cursor x position
    int _SavedCursY;       // Saved cursor y position

    int _BufChainNum;      // Buffer chain number
    int _WinLineNum;       // Window line number

    // Internal data monitoring function pointers
    const int*(*_insert_monitor)(void*, WindowRefLineInfo*, const int*);    // internal use
    const int*(*_key_monitor)(void*, WindowRefLineInfo*, const int*);
    const int*(*_ext_monitor)(void*, WindowRefLineInfo*, const int*);
    const int*(*_index_monitor)(void*, WindowRefLineInfo*, const int*);
    const int*(*_escaped_monitor)(void*, WindowRefLineInfo*, const int*);
    const int*(*_hidden_monitor)(void*, WindowRefLineInfo*, const int*);
    const int*(*_active_monitor)(void*, WindowRefLineInfo*, const int*);
    const int*(*_infocus_monitor)(void*, WindowRefLineInfo*, const int*);
    const int*(*_inboxoffset_monitor)(void*, WindowLineBuffer*, const int*);
    const int*(*_winlinenum_monitor)(void*, WindowLineBuffer*, const int*);    

    // Handles to external objects, used by data monitors
    void* _insert_handle;
    void* _key_handle;
    void* _ext_handle;
    void* _index_handle;
    void* _escaped_handle;
    void* _hidden_handle;
    void* _active_handle;
    void* _infocus_handle;
    void* _inboxoffset_handle;
    void* _winlinenum_handle;

    // Monitor functions execution and management methods
    const char* cMonitor(const char*(*monptr_)(void*, WindowRefLineInfo*, const char*), void* This_, const char* ptr_);
    const int* iMonitor(const int*(*monptr_)(void*, WindowRefLineInfo*, const int*), void* This_, const int* ptr_);
    void SaveCursorPosition();
    void RestoreCursorPosition();

    const int* MonitorInsertMode(void* This_, const int*(*FnPtr_)(void*, WindowRefLineInfo*, const int*));

    // Promptbox chaining internal methods
    void SetChainPos(int Pos_);

    // Insert state methods
    void NotifyInsertMode(const int* ModePtr_);
    void SetInsertMode(int Mode_, WindowRefLineInfo* Parent_=NULL);
    inline int GetInsertMode()
        { return _InsertMode; }

    // show prompt box methods
    virtual void ShowMessage();
    virtual void ShowInput(int status_);
    virtual void ResetPrompt();

    // Clear prompt box method
    void PrintBlanks(int Max_);

    // input validation / testing methods
    int* WordChr(int* src_, int val_, size_t n);
    inline int IsValidString(char* str)
        { return 1; }
    inline int IsValidLength() const
        { return 1; }
    int IsUserKey(int key);
    int ShouldEnd(int key, int zerotozero);
    int IsValidCtrlKey(int key);

    // Extended box methods
    inline void SetEscaped(bool Flag_)
        { _Escaped = Flag_; }
    inline WindowLineBuffer::SharedChainData* SharedData()
        { return (_parent ? _parent->SharedData():NULL); }
    inline const WindowLineBuffer::SharedChainData* ConstSharedData() const
        { return (_parent ? _parent->ConstSharedData():NULL); }

    // console input retrieval methods
    int GetKey();
    inline int GetExtKey()
        { return _extkey; }
    int GetLine(int status_);
    int ReadInput(int status_);
    void ResetInputState(int status_);
    void SetEnterable(Boolean Flag_);
    void SetSelectable(Boolean Flag_);
#if !(defined(__TURBOC__) | defined(__BORLANDC__) | defined(__DJGPP__)) & (defined(__linux__) | defined(__unix__))
    int TranslateKeys(int key);
#endif

    // keyboard character output methods
    bool PutExtKey(int key);
    bool PutStdKey(int key);
    void CommitPendingMove();
    void SetMovePending(bool pending_, bool done_);
    bool MovePending() const;
    void SetMoveDone(bool move_);
    bool MoveDone() const;

    // String push/pull methods
    static void FillString(char* Str_, char FillCh_, int Start_, int End_);
    int HasNonWsChars(const char* str_) const;
    void ShiftDownAndSetStr(char* NewStr_);
    inline void ChainPush(char ch) {}
    inline void ChainJoin(char ch) {}
    inline char ChainPull()
        { return 0; }
    inline void Push(int start, char ch=' ') {}
    inline void Join(char ch) {}
    inline char Pull(int start)
        { return 0; }
    inline char Peek(int Index_=0)
        { return 0; }

    // data member destruction method
    virtual void Clear();

    // promptbox constructors
    WindowRefLineInfo(int BufChainNum_, WindowLineBuffer* Parent_,
                    ControlWindow* Pwin_, int Length_=0);
    WindowRefLineInfo(WindowRefLineInfo& pb, int Length_=0);

    // Text color methods
    void SetupColors(int Mode_);

  public:
    // WindowRefLineInfo Destructor
    virtual ~WindowRefLineInfo()
        { Clear(); }

    // Promptbox chaining method
    WindowRefLineInfo* Chain(WindowRefLineInfo* next_);
    inline void EraseChain()
        { return; }
    inline WindowLineBuffer* Parent()
        { return _parent; }
    inline WindowRefLineInfo* NextChain()
        { return NULL; }

    virtual int IsFullBox() const;
    virtual int RestrictFullBox() const;
    virtual bool EndBrksActive() const;
    virtual bool ShouldTrimWs() const;
    virtual bool ExtendedIndexToggled() const;
    virtual void ResetExtendedIndexToggle();
    
    virtual int BufferIndex() const;
    virtual int InBoxOffset() const;
    virtual int InWBoxCount() const;
    virtual int InputLength() const;
    virtual int ParentIndex() const;
    virtual int ExtendedIndex() const;
    virtual int BoxPos() const;
    virtual int BoxDiff() const;
    virtual int Distance() const;

    // promptbox assignment operator
    virtual WindowRefLineInfo& operator = (WindowRefLineInfo& pb);
    static WindowRefLineInfo* Make(int BufChainNum_, WindowLineBuffer* Parent_,
                                 ControlWindow* Pwin_, int Length_=0);
    static WindowRefLineInfo* Make(WindowRefLineInfo& pbox, int Length_=0);
    static WindowLineBuffer* MakeNextLine(WindowLineBuffer* wpb);

    // prompt box properties mutators
    virtual char* SetMessage(char* str);
    virtual void RemoveMessage();
    virtual void BlankMessage();
    virtual void SetInputString(char* str);
    virtual void SetInputStringAtPos(char* str, int len, int pos, bool endterm_);
    virtual void FillCharAtPos(char ch, int pos, bool endterm_);
    virtual void PushInputString(char* str);
    virtual void SetInputLength(int len, bool MakeVbuf_);
    virtual void DivBufSizeAsInputLength(int div_, bool MakeVbuf_);
    virtual void SetMinimumLength(int len);
    void SetNoEmpty(int flag=1);
    int NextLengthHelper() const;
    int TakeNextLastCharHelper();
    bool SetXYpos(int x, int y);
    bool SetXpos(int x);
    bool SetYpos(int y);
    int GetXpos() const;
    int GetYpos() const;

    // virtual memory buffer methods
    virtual int VirtualBufferOffset(int* EndReached_=NULL) const;    
    virtual int VirtualBufferMax() const;
    virtual int IsChainable(bool HasChars_) const;
    virtual int NextLength() const;
    virtual int TakeNextLastChar();
    virtual int IsDummyLine() const;
    virtual int IsWindowLineBuffer() const;
    virtual int ChainPosition() const;
    virtual int BufferChainPosition() const;

    // prompt box manipulators
    virtual void Show();
    virtual void ShowInput();
    virtual void Hide();
    virtual int Prompt(int status_);
    virtual void Erase();

    // Accelerator key methods
    virtual int GetHotKey() const;
    virtual int GetShortCut() const;
    virtual int GetHotKeyType() const;

    // text control selection/deselection mutators
    virtual void Select();
    virtual void Deselect();

    // Empty, valid and selected Text Control test
    virtual Boolean IsEnterable() const;
    virtual Boolean IsSelectable() const;
    virtual Boolean IsSelected() const;
    virtual Boolean IsEmpty() const;
    virtual Boolean IsEscaped() const;
    virtual Boolean IsValid() const;
    virtual Boolean IsActive() const;
    virtual Boolean IsInFocus() const;

    // Activation/Deactivation for this text widget
    virtual void Activate();
    virtual void Deactivate();

    // inbox index methods
    virtual void DecrBoxIndex();
    virtual void IncrBoxIndex();
    virtual void ResetBoxIndex();
    virtual void SetBoxIndex(int Val_);

    // input data retrieval method
    char LastKeyRead(int* Ext_=NULL) const;
    int LastExtKeyRead() const;
    char ChrRetrieve(int Index_);
    inline char* GetMessage()
        { return NULL; }
    inline char* StrRetrieve()
        { return NULL; }

    // Text control ownership methods
    virtual void ChainSetOwner(ControlWindow* Pwin_, TextControl* TxtCtrl_);
    virtual void SetOwner(ControlWindow* Pwin_, TextControl* TxtCtrl_);
    virtual Boolean IsOwner(ControlWindow* Pwin_) const;

    // Text color methods
    virtual void UseColor(Boolean Flag_=TRUE);
    virtual void SetNormalColors(int FgCol_, int BgCol_);
    virtual void SetHighLightColors(int FgCol_, int BgCol_);
    virtual void SetHotKeyColor(int Col_);
    virtual void SetBackGroundColor(int Col_);
    virtual void UseDefaultColors();

    virtual Boolean UsesColor() const;
    virtual int* NormalColors() const;
    virtual int* HighLightColors() const;    
    virtual int HotKeyColor() const;
    virtual int BackGroundColor() const;

    // WindowLineBufferType specific methods
    virtual void SetWindowLineNum(int Index_);
    virtual void SetBufferChainNum(WindowLineBufferType* Lbuf_, int Index_);
    virtual void CreateReturnBuffer(int Size_);
    virtual void SetReturnBuffer(WindowLineBufferType* Lbuf_,
                                 char* RetBufPtr_, bool Alloc_, int Size_=0);

    // Internal data monitoring methods
    const char* MonitorBuffer(void* This_, const char*(*FnPtr_)(void*, WindowRefLineInfo*, const char*));
    const int* MonitorKey(void* This_, const int*(*FnPtr_)(void*, WindowRefLineInfo*, const int*));
    const int* MonitorExtendedKey(void* This_, const int*(*FnPtr_)(void*, WindowRefLineInfo*, const int*));
    const int* MonitorIndex(void* This_, const int*(*FnPtr_)(void*, WindowRefLineInfo*, const int*));
    const int* MonitorEscapedStatus(void* This_, const int*(*FnPtr_)(void*, WindowRefLineInfo*, const int*));
    const int* MonitorHiddenStatus(void* This_, const int*(*FnPtr_)(void*, WindowRefLineInfo*, const int*));
    const int* MonitorActiveStatus(void* This_, const int*(*FnPtr_)(void*, WindowRefLineInfo*, const int*));
    const int* MonitorInFocusStatus(void* This_, const int*(*FnPtr_)(void*, WindowRefLineInfo*, const int*));
    const int* MonitorInboxOffset(void* This_, const int*(*FnPtr_)(void*, WindowLineBuffer*, const int*));
    const int* MonitorWinLineNum(void* This_, const int*(*FnPtr_)(void*, WindowLineBuffer*, const int*));
};

/****************************************************************************/
/****************************************************************************/
class WindowLineRef;

struct WinCoordsData
{
  bool _Allocated;   // memory allocation flag

  // text mode screen dimensions:
  //   _screenmaxx == TextGraphics::TerminalScreenCols()
  //   _screenmaxy == TextGraphics::TerminalScreenRows()
  //
  //   1 <= _screenxpos <= _screenmaxx
  //   1 <= _screenypox <= _screenmaxy
  //
  //   _screenxpos : determined from cursor pos.
  //   _screenypos : determined from cursor pos.
  //   _screenmaxx : assigned from TextGraphics::TerminalScreenCols()
  //   _screenmaxy : assigned from TextGraphics::TerminalScreenRows()
  //
  int _screenxpos;   // horizontal coordinate position of physical text screen
  int _screenypos;   // vertical coordinate position of physical text screen
  int _screenmaxx;   // maximum horizontal extent of physical text screen
  int _screenmaxy;   // maximum vertical extent of physical text screen

  // subset coordinates of enclosing screen dimensions:
  //   1 <= _winboxstartx <= _screenmaxx
  //   1 <= _winboxstarty <= _screenmaxy
  //
  //   _winboxstartx <= _winboxxpos <= _winboxmaxx
  //   _winboxstarty <= _winboxypos <= _winboxmaxy
  //
  //   WinBoxLength() == _winboxmaxx - _winboxstartx + 1
  //   WinBoxHeight() == _winboxmaxy - _winboxstarty + 1
  //
  //   _winboxxpos   : determined from cursor pos.
  //   _winboxypos   : determined from cursor pos.
  //   _winboxstartx : passed from method or constructor argument
  //   _winboxstarty : passed from method or constructor argument
  //   _winboxmaxx   : passed from method or constructor argument
  //   _winboxmaxy   : passed from method or constructor argument
  //
  int _winboxxpos;   // horizontal coordinate position of bounded windows box
  int _winboxypos;   // vertical coordinate position of bounded windows box
  int _winboxstartx; // starting x of bounded windows box in text screen coordinates
  int _winboxstarty; // starting y of bounded windows box in text screen coordinates
  int _winboxmaxx;   // maximum horizontal extent of bounded windows box
  int _winboxmaxy;   // maximum vertical extent of bounded windows box

  // subset coordinates of enclosing window box:
  //   1 <= _winviewxpos <= _winviewmaxx - _winboxstartx + 1
  //   1 <= _winviewypos <= _winviewmaxy - _winboxstarty + 1
  //
  //   _winviewstartx same as _winboxstartx
  //   _winviewstarty same as _winboxstarty
  //
  //   WinViewLength() == _winviewmaxx - _winboxstartx + 1
  //   WinViewHeight() == _winviewmaxy - _winboxstarty + 1
  //
  //   if _winviewmaxx > _winboxmaxx
  //     1 <= _winviewxpos - _winviewxoffset <= WinBoxLength()
  //
  //   if _winviewmaxy > _winboxmaxy
  //     1 <= _winviewypos - _winviewyoffset <= WinBoxHeight()
  //
  //    _winviewxoffset : determined from cursor pos.
  //    _winviewyoffset : determined from cursor pos.
  //    _winviewxpos    : determined from cursor pos.
  //    _winviewypos    : determined from cursor pos.
  //    _winviewmaxx    : passed from method argument
  //    _winviewmaxy    : passed from method argument
  //
  int _winviewxoffset; // shifted x offset of sliding windows view
  int _winviewyoffset; // shifted y offset of sliding window view
  int _winviewxpos;    // horizontal coordinate position of sliding windows view
  int _winviewypos;    // vertical coordinate position of sliding window view
  int _winviewmaxx;    // maximum horizontal extent of sliding windows view
  int _winviewmaxy;    // maximum vertical extent of sliding window view

  WinCoordsData(int WinStartx_, int WinStarty_,
                int WinMaxx_, int WinMaxy_);
  WinCoordsData(const WinCoordsData& Obj_);

  static WinCoordsData* Make(int WinStartx_, int WinStarty_,
                             int WinMaxx_, int WinMaxy_);
  static WinCoordsData* Make(const WinCoordsData& Obj_);

  WinCoordsData& operator = (const WinCoordsData& Obj_);

  void CopyCursorPosition(const WinCoordsData& Obj_);
  void SetWinBoxDimensions(int WinStartx_, int WinStarty_,
                           int WinMaxx_, int WinMaxy_);
  void SetWinViewMaxXY(int maxx_, int maxy_);

  int WinViewLength() const;
  int WinViewHeight() const;

  int WinBoxLength() const;
  int WinBoxHeight() const;
};

class WindowLine : public ControlWindow
{
  friend class WindowLineRef;
  friend class WindowLineBuffer;
  friend class WindowRefLineInfo;

  protected:
    struct SharedChainData
    {
      WindowLine* _parent;
      WindowLine* _currentrow;
      WindowLine* _rowcursor;
    
      bool _InWinMode;           // in text editor window mode
      bool _PrevLineReady;       // previous text line available in text editor window
      bool _NextLineReady;       // next text line available in text editor window
      bool _ReturnOnLastLine;    // Inactivate WindowLine upon ENTER key on last line
      bool _LastLineTerminated;  // Last line on window block terminated
      bool _AddNewLinesWhenFull; // Add new lines when block is all filled
      bool _WinLinePosTransfer;  // Window line position tranferring in process

      bool _rowcursorfound;    // line position saved
      bool _movedone;          // key move done for window line
      bool _rowpossaved;       // line position saved
      bool _hasavailemptyline; // has available empty line
      int _numnextlines;       // number of nextline buffers
      int _rowpos;             // line position (row position in window block)
      int _savedrowpos;        // saved line position
      int _availlineno;        // available line number
      int _newlineslimit;      // upper limit on number of new lines to add
                               // or infinite if equal to -1

      SharedChainData(WindowLine* wlptr=NULL);

      void SetParent(WindowLine* wlptr);
    };

    enum
    {
      UNINIT_NL = -1,
      JOINED_NL = 0,
      INIT_NL   = 1,

      TO_PARENT  = 100,
      TO_CURRENT = 101,
      GROW_INCR = 5
    };
  
    TextControl** _pbox;      // prompt box array
    ControlInfo* _Info;
    WindowLine* _nextline;    // next window line in window block
    WindowLine* _prevline;    // back pointer to previous line in DblLnk list
    WindowLineRef** _RefPtr;
    WindowLineRef* _DummyRef;

    bool _SharedDataErased;
    SharedChainData* _SharedData;
    
    bool _LineTerminated;   // Winbuffer series terminated
    bool _LineWrapAllowed;  // Line wrapping is allowed
    bool _ShowChainedBoxes; // Show chained buffer boxes.
    bool _WindowLineActive; // Active flag for this window line object
    bool _EmptyLine;        // Indicates blank line with no content

    int _linestatus;        // line status : { UNINIT, JOINED, INIT }
    int _maxbox;            // maximum number of prompt boxes
    int _boxcount;          // total non-null pointers to windows buffers stored

    WinCoordsData* _WinCoords;

    int _inputlen;        // maximum input length (should set to same for all buffers)
    int _vbuflen;         // virtual buffer length for variable length single line input
    int _maxwraparound;   // maximum wrap around for window buffer contents
    int _wraparoundindex; // running wrap around index count
    int _WindowLineNum;   // Window line number for this line object

    // Extended window lines methods
    bool EndLineEmpty();
    bool AllLineFilled();
    void ResetLineData(bool ResetRow_);
    bool IsAtEndPt();
    void EraseSharedChainData(WindowLine::SharedChainData* ptr);
    void FindLineStatus();
    inline int BoxInputLength() const
        { return _inputlen; }

    // Dynamic window lines chaining methods
    bool AtLineChainLimit();
    WindowLine* LastChainLine();
    WindowLine* IncrLineChains(int NumWinLines_=0);
    WindowLine* SetupChainWlineObject(WindowLine* ptr, bool ChainPb_);
    WindowLineBuffer* SetupFirstBox(WindowLine& pw, int& VbufLen_);
    void SetEscaped(bool Flag_);
    void Activate();
    void Deactivate();
    void ReCalcChainWline(bool ReCalcSdp_, bool ReCalcStatus_);
    void SaveLinePos();
    WindowLine* RestoreLinePos();
    WindowLine* DetachNext(WindowLine* Chainp_);
    WindowLine* DetachThis(WindowLine* Chainp_);
    bool RemoveLineChains(WindowLine* wlptr);
    bool WindowLineActive() const
        { return _WindowLineActive; }

    // row position cursor methods
    void SetRowCursor(WindowLine* ptr, bool AsCurrent_);
    void InitRowCursor();
    void RestoreRowCursor(int DestPos_);

    // Text Control accessor method
    virtual void Control(int Index_, TextControl* TxtCtrl_);
    virtual TextControl* Control(int Index_);

    // Dialog controls validity check method
    virtual Boolean AnyActive();
    virtual Boolean ValidityCheck();

    // data member destruction method
    virtual void Clear();

    // convenience methods    
    TextControl** NewPboxes(size_t max, bool classdata_=true);
    WindowLineRef** NewRefPtrs(size_t max, bool classdata_=true);
    void EraseBox(int BoxIndex_);
    void ReplaceBox(int BoxIndex_, TextControl* ptr_);
    void InitializeBox(WindowLineBuffer* ptr, int x,
                       bool CopyBox_, bool JoinBoxToEnd_);
    WindowLine* CopyLineInfo(WindowLine* ptr_, bool DeepCopy_, int Length_=0);
    void EraseLineInfo(bool MemDelete_);
    void SetControlData(TextControl** CtrlArray_, ControlInfo* CtrlInfo_,
                        int index, bool CheckNextLinks_=true);
    void RemoveRefPtrData(WindowLineRef** RefPtrArray_, int index);
    void RemoveControlData(TextControl** CtrlArray_, ControlInfo* CtrlInfo_,
                           int index, bool CheckNextLinks_=true);
    void ReplaceControlData(TextControl** CtrlArray_, ControlInfo* CtrlInfo_,
                            int index, bool CheckNextLinks_=true);
    void LinkNextControl(TextControl** CtrlArray_, int index);
    void UnlinkNextControl(TextControl** CtrlArray_, int index);
    bool HasControlArrays(WindowLine* Pwin_);
    bool NoMatchesToPrevBoxes(WindowLineBuffer* Nbox_, WindowLineBuffer* PrevBox_,
                              TextControl** Pbox_, int Maxbox_);

    // WindowLine class specific methods
    int FindMaxWrapAround();
    void SetBoxToString(char* NewStr_, int index);

    // Chain fill all following boxes with specified fill character
    void ChainFillString(char NewCh_, int Offset_,
                         TextControl** Pbox_, int Boxdex_, int Maxbox_);
    void ChainFillString(char* NewStr_, int Length_, int Offset_,
                         TextControl** Pbox_, int Boxdex_, int Maxbox_);

  public:
    // prompt window constructors
    WindowLine(int Max_, int InputLen_, int VbufLen_,
               WinCoordsData* CoordData_);
    WindowLine(WindowLine& pw);
    virtual ~WindowLine();

    static WindowLine* Make(int Max_, int InputLen_, int VbufLen_,
                            WinCoordsData* CoordData_);
    static WindowLine* Make(WindowLine& wline_);
    static WindowLine* MakeNextLine(WindowLine* wl);
    virtual WindowLine& operator = (WindowLine& wline_);

    // prompt window data mutators
    virtual void MaxControl(int max);
    virtual void Grow(int extra);

    // data member accessors
    virtual int MaxControl();
    virtual void ReplaceControl(TextControl* TxtCtrl_);
    virtual void RemoveControl(TextControl* TxtCtrl_);
    virtual void DeleteControl(TextControl* TxtCtrl_);
    virtual int FindControlCount();

    // Promptbox specific accessors
    WindowLineRef& operator [] (int index);
    WindowLineBuffer* NextSelected(int& Start_);

    // prompt window manipulators
    virtual void Show();
    virtual void Hide();
    virtual int Prompt(int StartIndex_=0);
    virtual void Erase();
    virtual void Delete();
    virtual void Remove();

    void ShowWindowLineAt(int LineNum_, int Ypos_, int Xpos_);
    void ShowAll();
    void HideAll();
    void EraseAll();
    void DeleteAll();
    void RemoveAll();

    // Text color methods
    virtual void UseColor(Boolean Flag_=TRUE);
    virtual void SetNormalColors(int FgCol_, int BgCol_);
    virtual void SetHighLightColors(int FgCol_, int BgCol_);
    virtual void SetHotKeyColor(int Col_);
    virtual void UseDefaultColors();

    // Empty, valid and selected Text Control test
    virtual Boolean IsSelected() const;
    virtual Boolean IsEmpty() const;
    virtual Boolean IsEscaped() const;
    virtual Boolean IsValid() const;
    virtual Boolean IsActive() const;
    virtual Boolean IsInFocus() const;

    // Retrieve control information object
    virtual ControlInfo* GetControlInfo();

    // new WindowLine specific methods
    WindowLine& SetWinLineProperties(int LwAllowed_=-99,
                                     int ShowChained_=-99,
                                     int RestrictFull_=-99,
                                     int TrimWs_=-99,
                                     int EndBrksActive_=-99);
    WindowLine& SetWinBlockProperties(int InWin_=-99,
                                      int RetOnLast_=-99,
                                      int AddNewLines_=-99,
                                      int NewLinesLimit_=-99);
    void SendUnprocessedKey(int Key_, int Extended_,
                            WindowLineBuffer* WinBufPtr_);
    WindowLineBuffer* ChainPbox(WindowLineBuffer* WinBufPtr_, int PboxNum_);
    WindowLine* ChainWline(WindowLine* next_, bool IncrNumChains_);

    bool AddNewWindowLines(int NewLines_=0);
    int SetInputLength(int Length_);
    void IncrWrapAround(int Num_);
    void DecrWrapAround(int Num_);
    bool InWindowMode(bool* HaveNext_=NULL, bool* HavePrev_=NULL);
    TextControl* ConfirmControlPtr(TextControl* ctrlptr_,
                                   int& index_, int* changed_=NULL);

    char* RetrieveInOrder();
    int GetWrapAroundIndex();

    void SetWindowLineNum(WindowLine* Wline_, int Index_);
    inline int GetMaxWrapAround()
        { return FindMaxWrapAround(); }
    inline bool LineWrapAllowed() const
        { return _LineWrapAllowed; }
    inline bool LineTerminated() const
        { return _LineTerminated; }
    bool AddNewLinesWhenFull() const;
    bool ReturnOnLastLine() const;
    inline bool ShowChainedBoxes() const
        { return _ShowChainedBoxes; }
    int GetBoxCount(bool Contiguous_=false);
    inline void ResetLine()
        { _LineTerminated = false; }
    inline int WindowLineNum() const
        { return _WindowLineNum; }

    int WinBoxToScreenX(int xpos_=-1);
    int WinBoxToScreenY(int ypos_=-1);
    int WinViewToWinBoxX(int xpos_=-1);
    int WinViewToWinBoxY(int ypos_=-1);

    WindowLine& GotoWindowLine(int linenum_);
    const WindowLine& GotoWindowLine(int linenum_) const;

    WindowLine& GotoNextWindowLine(bool* done_=NULL);
    const WindowLine& GotoNextWindowLine(bool* done_=NULL) const;
    WindowLine& GotoPrevWindowLine(bool* done_=NULL);
    const WindowLine& GotoPrevWindowLine(bool* done_=NULL) const;

    inline WindowLine* NextLinePtr()
        { return _nextline; }
    inline const WindowLine* NextLinePtr() const
        { return _nextline; }
    inline WindowLine* PreviousLinePtr()
        { return _prevline; }
    inline const WindowLine* PreviousLinePtr() const
        { return _prevline; }

    bool SetXYpos(int x, int y);
    bool SetXpos(int x);
    bool SetYpos(int y);
    int GetXpos() const;
    int GetYpos() const;

    // Shared chain data methods
    void AssignSharedData(bool& InWinMode_,
                          bool& NextLineReady_, bool& PrevLineReady);
    inline WindowLine::SharedChainData* SharedData()
        { return ((!_SharedDataErased && _SharedData) ? _SharedData:NULL); }
    inline const WindowLine::SharedChainData* ConstSharedData() const
        { return ((!_SharedDataErased && _SharedData) ? _SharedData:NULL); }

    // Transferring box line position from current line to next/prev line
    void TransferBoxPosToNextLine(int linedir_);

    // Perform shift down of entered line to next WindowLine object
    bool ShiftDown(char* NewStr_, int Offset_,
                   TextControl** Pbox_, int Boxdex_, int Maxbox_);
    bool ShiftDownNext(char* NewStr_, int Offset_,
                       TextControl** Pbox_, int Boxdex_, int Maxbox_);

    // Perform shift up of deleted line to current WindowLine object
    bool ShiftUp(char* NewStr_, int Offset_,
                 TextControl** Pbox_, int Boxdex_, int Maxbox_);
    bool ShiftUpNext(char* NewStr_, int Offset_,
                     TextControl** Pbox_, int Boxdex_, int Maxbox_);
};

/****************************************************************************/
class WindowLineRef
{
  friend class WindowLine;
  private:
    WindowLine* _BossPtr;
    int _Index;
    bool _Cleared;
    bool _InProcess;

    WindowLineRef(WindowLine* Boss_, int Index_);
    void ResetLineRef(WindowLine* Boss_, int Index_);

    inline bool InProcess() const
        { return _InProcess; }
    inline bool Cleared() const
        { return (_Cleared && !_InProcess); }
    static WindowLineBuffer* CastToWindowLineBuffer(TextControl* TxtCtrl_)
        { return ((WindowLineBuffer*)TxtCtrl_); }

  public:
    WindowLineRef& operator = (WindowLineBuffer* Pbox_);
    operator WindowLineBuffer* ();
    WindowLineBuffer* operator -> ();
};

/****************************************************************************/
class WindowBlock
{
  public:
    // prompt window constructors
    WindowBlock(WindowLine* Line_, int NumLines_);
    WindowBlock(WindowBlock& pwb);
    virtual ~WindowBlock();
  
    // prompt window manipulators
    virtual void Show();
    virtual void Hide();
    virtual int Prompt(int StartIndex_=0);
    virtual void Erase();
    virtual void Delete();
    virtual void Remove();

    void ShowAll();
    void HideAll();
    void EraseAll();
    void DeleteAll();
    void RemoveAll();
};

/****************************************************************************/
#endif
